@@ -2,14 +2,7 @@
 #define _BASE_H_
 #include "first.h"
 
-#include <sys/types.h>
-#include <sys/time.h>
-
-#if defined(__APPLE__) && defined(__MACH__)
-#if __ENVIRONMENT_MAC_OS_X_VERSION_MIN_REQUIRED__ < 1050
-#include <time.h>
-#endif
-#endif
+#include "sys-time.h"
 
 #include "base_decls.h"
 #include "buffer.h"
@@ -28,8 +21,8 @@ struct connection {
 	h2con *h2;
 
 	int fd;                      /* the FD for this connection */
-	int ndx;                     /* reverse mapping to server->connection[ndx] */
 	fdnode *fdn;                 /* fdevent (fdnode *) object */
+	connection *jqnext;
 
 	/* fd states */
 	signed char is_readable;
@@ -56,19 +49,34 @@ struct connection {
 	void *config_data_base;
 
 	sock_addr dst_addr;
-	buffer *dst_addr_buf;
-	struct server_socket *srv_socket;   /* reference to the server-socket */
+	buffer dst_addr_buf;
+	const struct server_socket *srv_socket;   /* reference to the server-socket */
 
 	/* timestamps */
-	time_t read_idle_ts;
-	time_t close_timeout_ts;
-	time_t write_request_ts;
+	unix_time64_t read_idle_ts;
+	unix_time64_t close_timeout_ts;
+	unix_time64_t write_request_ts;
+	unix_time64_t connection_start;
 
-	time_t connection_start;
 	uint32_t request_count;      /* number of requests handled in this connection */
 	int keep_alive_idle;         /* remember max_keep_alive_idle from config */
+
+	connection *next;
+	connection *prev;
 };
 
+/* log_con_jqueue is in log.c to be defined in shared object */
+#define joblist_append(con) connection_jq_append(con)
+extern connection *log_con_jqueue;
+static inline void connection_jq_append(connection * const restrict con);
+static inline void connection_jq_append(connection * const restrict con)
+{
+    if (!con->jqnext) {
+        con->jqnext = log_con_jqueue;
+        log_con_jqueue = con;
+    }
+}
+
 typedef struct {
 	connection **ptr;
 	uint32_t size;
@@ -131,7 +139,8 @@ typedef struct server_socket {
 	sock_addr addr;
 	int       fd;
 
-	unsigned short is_ssl;
+	uint8_t is_ssl;
+	uint8_t srv_token_colon;
 	unsigned short sidx;
 
 	fdnode *fdn;
@@ -157,11 +166,6 @@ struct server {
 	/* buffers */
 	buffer *tmp_buf;
 
-	connections conns;
-	connections joblist_A;
-	connections joblist_B;
-	connections fdwaitqueue;
-
 	/* counters */
 	int con_opened;
 	int con_read;
@@ -174,11 +178,13 @@ struct server {
 	int cur_fds;    /* currently used fds */
 	int sockets_disabled;
 
-	uint32_t max_conns;
+	uint32_t lim_conns;
+	connection *conns;
+	connection *conns_pool;
 
 	log_error_st *errh;
 
-	time_t loadts;
+	unix_time64_t loadts;
 	double loadavg[3];
 
 	/* members used at start-up or rarely used */
@@ -190,8 +196,8 @@ struct server {
 	server_socket_array srv_sockets_inherited;
 	buffer_plugin plugins;
 
-	time_t startup_ts;
-	time_t graceful_expire_ts;
+	unix_time64_t startup_ts;
+	unix_time64_t graceful_expire_ts;
 
 	uid_t uid;
 	gid_t gid;
