@@ -24,17 +24,16 @@ typedef union sock_addr sock_addr;
 struct fdnode_st;
 typedef struct fdnode_st fdnode;
 
-struct log_error_st;
-typedef struct log_error_st log_error_st;
+struct fdlog_st;
+typedef struct fdlog_st fdlog_st;
+typedef struct fdlog_st log_error_st;
 
 enum handler_t {
-  HANDLER_UNSET,
   HANDLER_GO_ON,
   HANDLER_FINISHED,
   HANDLER_COMEBACK,
   HANDLER_WAIT_FOR_EVENT,
-  HANDLER_ERROR,
-  HANDLER_WAIT_FOR_FD
+  HANDLER_ERROR
 };
 typedef enum handler_t handler_t;
 
