@@ -39,16 +39,34 @@ void config_get_config_cond_info(config_cond_info * const cfginfo, uint32_t idx)
     cfginfo->comp_key = dc->comp_key;
 }
 
+int config_feature_bool (const server *srv, const char *feature, int default_value) {
+    return srv->srvconf.feature_flags
+      ? config_plugin_value_tobool(
+          array_get_element_klen(srv->srvconf.feature_flags,
+                                 feature, strlen(feature)), default_value)
+      : default_value;
+}
+
+int32_t config_feature_int (const server *srv, const char *feature, int32_t default_value) {
+    return srv->srvconf.feature_flags
+      ? config_plugin_value_to_int32(
+          array_get_element_klen(srv->srvconf.feature_flags,
+                                 feature, strlen(feature)), default_value)
+      : default_value;
+}
+
 int config_plugin_value_tobool (const data_unset *du, int default_value)
 {
     if (NULL == du) return default_value;
     if (du->type == TYPE_STRING) {
         const buffer *b = &((const data_string *)du)->value;
         if (buffer_eq_icase_slen(b, CONST_STR_LEN("enable"))
+            || buffer_eq_icase_slen(b, CONST_STR_LEN("enabled"))
             || buffer_eq_icase_slen(b, CONST_STR_LEN("true"))
             || buffer_eq_icase_slen(b, CONST_STR_LEN("1")))
             return 1;
         else if (buffer_eq_icase_slen(b, CONST_STR_LEN("disable"))
+                 || buffer_eq_icase_slen(b, CONST_STR_LEN("disabled"))
                  || buffer_eq_icase_slen(b, CONST_STR_LEN("false"))
                  || buffer_eq_icase_slen(b, CONST_STR_LEN("0")))
             return 0;
@@ -355,15 +373,6 @@ static cond_result_t config_check_cond_nocache(request_st *r, const data_config
 
 static cond_result_t config_check_cond_nocache_calc(request_st * const r, const data_config * const dc, const int debug_cond, cond_cache_t * const cache) {
     cache->result = config_check_cond_nocache(r, dc, debug_cond, cache);
-    switch (cache->result) {
-      case COND_RESULT_FALSE:
-      case COND_RESULT_TRUE:
-        /* remember result of local condition for a partial reset */
-        cache->local_result = cache->result;
-        break;
-      default:
-        break;
-    }
     if (debug_cond) config_cond_result_trace(r, dc, 0);
     return cache->result;
 }
@@ -377,66 +386,9 @@ static cond_result_t config_check_cond_cached(request_st * const r, const data_c
     return config_check_cond_nocache_calc(r, dc, debug_cond, cache);
 }
 
-static int config_addrstr_eq_remote_ip_mask(request_st * const r, const char * const addrstr, const int nm_bits, sock_addr * const rmt) {
-	/* special-case 0 == nm_bits to mean "all bits of the address" in addrstr */
-	sock_addr addr;
-	if (1 == sock_addr_inet_pton(&addr, addrstr, AF_INET, 0)) {
-		if (nm_bits > 32) {
-			log_error(r->conf.errh, __FILE__, __LINE__, "ERROR: ipv4 netmask too large: %d", nm_bits);
-			return -1;
-		}
-	} else if (1 == sock_addr_inet_pton(&addr, addrstr, AF_INET6, 0)) {
-		if (nm_bits > 128) {
-			log_error(r->conf.errh, __FILE__, __LINE__, "ERROR: ipv6 netmask too large: %d", nm_bits);
-			return -1;
-		}
-	} else {
-		log_error(r->conf.errh, __FILE__, __LINE__, "ERROR: ip addr is invalid: %s", addrstr);
-		return -1;
-	}
-	return sock_addr_is_addr_eq_bits(&addr, rmt, nm_bits);
-}
-
-static int config_addrbuf_eq_remote_ip_mask(request_st * const r, const buffer * const string, char * const nm_slash, sock_addr * const rmt) {
-	char *err;
-	int nm_bits = strtol(nm_slash + 1, &err, 10);
-	size_t addrstrlen = (size_t)(nm_slash - string->ptr);
-	char addrstr[64]; /*(larger than INET_ADDRSTRLEN and INET6_ADDRSTRLEN)*/
-
-	if (*err) {
-		log_error(r->conf.errh, __FILE__, __LINE__, "ERROR: non-digit found in netmask: %s %s", string->ptr, err);
-		return -1;
-	}
-
-	if (nm_bits <= 0) {
-		if (*(nm_slash+1) == '\0') {
-			log_error(r->conf.errh, __FILE__, __LINE__, "ERROR: no number after / %s", string->ptr);
-		} else {
-			log_error(r->conf.errh, __FILE__, __LINE__, "ERROR: invalid netmask <= 0: %s %s", string->ptr, err);
-		}
-		return -1;
-	}
-
-	if (addrstrlen >= sizeof(addrstr)) {
-		log_error(r->conf.errh, __FILE__, __LINE__, "ERROR: address string too long: %s", string->ptr);
-		return -1;
-	}
-
-	memcpy(addrstr, string->ptr, addrstrlen);
-	addrstr[addrstrlen] = '\0';
-
-	return config_addrstr_eq_remote_ip_mask(r, addrstr, nm_bits, rmt);
-}
-
 static int data_config_pcre_exec(const data_config *dc, cond_cache_t *cache, const buffer *b, cond_match_t *cond_match);
 
 static cond_result_t config_check_cond_nocache(request_st * const r, const data_config * const dc, const int debug_cond, cond_cache_t * const cache) {
-	static struct const_char_buffer {
-	  const char *ptr;
-	  uint32_t used;
-	  uint32_t size;
-	} empty_string = { "", 1, 0 };
-
 	/* check parent first */
 	if (dc->parent && dc->parent->context_ndx) {
 		/**
@@ -504,138 +456,104 @@ static cond_result_t config_check_cond_nocache(request_st * const r, const data_
 		break;
 	}
 
-	if (CONFIG_COND_ELSE == dc->cond) return COND_RESULT_TRUE;
+	if (CONFIG_COND_ELSE == dc->cond)
+		return (cache->local_result = COND_RESULT_TRUE);
+		/* remember result of local condition for a partial reset */
 
 	/* pass the rules */
 
-	buffer *l;
+	static struct const_char_buffer {
+	  const char *ptr;
+	  uint32_t used;
+	  uint32_t size;
+	} empty_string = { "", 1, 0 };
+
+	const buffer *l;
 	switch (dc->comp) {
 	case COMP_HTTP_HOST:
-
 		l = &r->uri.authority;
-
-		if (buffer_string_is_empty(l)) {
-			l = (buffer *)&empty_string;
-			break;
-		}
-
-		switch(dc->cond) {
-		case CONFIG_COND_NE:
-		case CONFIG_COND_EQ: {
-			unsigned short port = sock_addr_get_port(&r->con->srv_socket->addr);
-			if (0 == port) break;
-			const char *ck_colon = strchr(dc->string.ptr, ':');
-			const char *val_colon = strchr(l->ptr, ':');
-
-			/* append server-port if necessary */
-			if (NULL != ck_colon && NULL == val_colon) {
-				/* condition "host:port" but client send "host" */
-				buffer *tb = r->tmp_buf;
-				buffer_copy_buffer(tb, l);
-				buffer_append_string_len(tb, CONST_STR_LEN(":"));
-				buffer_append_int(tb, port);
-				l = tb;
-			} else if (NULL != val_colon && NULL == ck_colon) {
-				/* condition "host" but client send "host:port" */
-				buffer *tb = r->tmp_buf;
-				buffer_copy_string_len(tb, l->ptr, val_colon - l->ptr);
-				l = tb;
-			}
-			break;
-		}
-		default:
-			break;
-		}
-
 		break;
-	case COMP_HTTP_REMOTE_IP: {
-		char *nm_slash;
-		/* handle remoteip limitations
-		 *
-		 * "10.0.0.1" is provided for all comparisons
-		 *
-		 * only for == and != we support
-		 *
-		 * "10.0.0.1/24"
-		 */
-
-		if ((dc->cond == CONFIG_COND_EQ ||
-		     dc->cond == CONFIG_COND_NE) &&
-		    (NULL != (nm_slash = strchr(dc->string.ptr, '/')))) {
-			switch (config_addrbuf_eq_remote_ip_mask(r, &dc->string, nm_slash, &r->con->dst_addr)) {
-			case  1: return (dc->cond == CONFIG_COND_EQ) ? COND_RESULT_TRUE : COND_RESULT_FALSE;
-			case  0: return (dc->cond == CONFIG_COND_EQ) ? COND_RESULT_FALSE : COND_RESULT_TRUE;
-			case -1: return COND_RESULT_FALSE; /*(error parsing configfile entry)*/
-			}
-		}
-		l = r->con->dst_addr_buf;
+	case COMP_HTTP_REMOTE_IP:
+		l = &r->con->dst_addr_buf;
 		break;
-	}
 	case COMP_HTTP_SCHEME:
 		l = &r->uri.scheme;
 		break;
-
 	case COMP_HTTP_URL:
 		l = &r->uri.path;
 		break;
-
 	case COMP_HTTP_QUERY_STRING:
 		l = &r->uri.query;
-		if (NULL == l->ptr) l = (buffer *)&empty_string;
 		break;
-
 	case COMP_SERVER_SOCKET:
 		l = r->con->srv_socket->srv_token;
 		break;
-
 	case COMP_HTTP_REQUEST_HEADER:
-		*((const buffer **)&l) =
-		  http_header_request_get(r, dc->ext, CONST_BUF_LEN(&dc->comp_tag));
+		l = http_header_request_get(r, dc->ext, BUF_PTR_LEN(&dc->comp_tag));
 		if (NULL == l) l = (buffer *)&empty_string;
 		break;
 	case COMP_HTTP_REQUEST_METHOD:
-		l = r->tmp_buf;
-		buffer_clear(l);
-		http_method_append(l, r->http_method);
+		l = http_method_buf(r->http_method);
 		break;
 	default:
-		return COND_RESULT_FALSE;
+		return (cache->local_result = COND_RESULT_FALSE);
 	}
 
-	if (NULL == l) { /*(should not happen)*/
-		log_error(r->conf.errh, __FILE__, __LINE__,
-			"%s compare to NULL", dc->comp_key);
-		return COND_RESULT_FALSE;
-	}
-	else if (debug_cond) {
+	if (__builtin_expect( (buffer_is_blank(l)), 0))
+		l = (buffer *)&empty_string;
+
+	if (debug_cond)
 		log_error(r->conf.errh, __FILE__, __LINE__,
 			"%s compare to %s", dc->comp_key, l->ptr);
-	}
 
+	int match;
 	switch(dc->cond) {
 	case CONFIG_COND_NE:
 	case CONFIG_COND_EQ:
-		if (buffer_is_equal(l, &dc->string)) {
-			return (dc->cond == CONFIG_COND_EQ) ? COND_RESULT_TRUE : COND_RESULT_FALSE;
-		} else {
-			return (dc->cond == CONFIG_COND_EQ) ? COND_RESULT_FALSE : COND_RESULT_TRUE;
+		match = (dc->cond == CONFIG_COND_EQ);
+		if (dc->comp == COMP_HTTP_HOST && dc->string.ptr[0] != '/') {
+			uint_fast32_t llen = buffer_clen(l);
+			uint_fast32_t dlen = buffer_clen(&dc->string);
+			/* check names match, whether or not :port suffix present */
+			/*(not strictly checking for port match for alt-svc flexibility,
+			 * though if strings are same length, port is checked for match)*/
+			/*(r->uri.authority not strictly checked here for excess ':')*/
+			/*(r->uri.authority lowercased during request parsing)*/
+			if (llen && llen != dlen) {
+				match ^= ((llen > dlen)
+				             ? l->ptr[dlen] == ':' && llen - dlen <= 5
+				             : dc->string.ptr[(dlen = llen)] == ':')
+				         && 0 == memcmp(l->ptr, dc->string.ptr, dlen);
+				break;
+			}
 		}
+		else if (dc->comp == COMP_HTTP_REMOTE_IP && dc->string.ptr[0] != '/') {
+			/* CIDR mask comparisons only supported for COND_EQ, COND_NE */
+			/* compare using structure data after end of string
+			 * (generated at startup when parsing config) */
+			const sock_addr * const addr = (sock_addr *)
+			  (((uintptr_t)dc->string.ptr + dc->string.used + 1 + 7) & ~7);
+			int bits = ((unsigned char *)dc->string.ptr)[dc->string.used];
+			match ^= (bits)
+			  ? sock_addr_is_addr_eq_bits(addr, &r->con->dst_addr, bits)
+			  : sock_addr_is_addr_eq(addr, &r->con->dst_addr);
+			break;
+		}
+		match ^= (buffer_is_equal(l, &dc->string));
+		break;
 	case CONFIG_COND_NOMATCH:
 	case CONFIG_COND_MATCH: {
 		cond_match_t *cond_match = r->cond_match + dc->context_ndx;
-		if (data_config_pcre_exec(dc, cache, l, cond_match) > 0) {
-			return (dc->cond == CONFIG_COND_MATCH) ? COND_RESULT_TRUE : COND_RESULT_FALSE;
-		} else {
-			/* cache is already cleared */
-			return (dc->cond == CONFIG_COND_MATCH) ? COND_RESULT_FALSE : COND_RESULT_TRUE;
-		}
-	}
-	default:
-		/* no way */
+		match = (dc->cond == CONFIG_COND_MATCH);
+		match ^= (data_config_pcre_exec(dc, cache, l, cond_match) > 0);
 		break;
 	}
-
-	return COND_RESULT_FALSE;
+	default:
+		match = 1; /* return (cache->local_result = COND_RESULT_FALSE); below */
+	}
+	/* remember result of local condition for a partial reset */
+	cache->local_result = match ? COND_RESULT_FALSE : COND_RESULT_TRUE;
+	return cache->local_result;
 }
 
 __attribute_noinline__
@@ -718,7 +636,7 @@ static int data_config_pcre_exec(const data_config *dc, cond_cache_t *cache, con
     #define elementsof(x) (sizeof(x) / sizeof(x[0]))
     #endif
     cache->patterncount =
-      pcre_exec(dc->regex, dc->regex_study, CONST_BUF_LEN(b), 0, 0,
+      pcre_exec(dc->regex, dc->regex_study, BUF_PTR_LEN(b), 0, 0,
                 cond_match->matches, elementsof(cond_match->matches));
     if (cache->patterncount > 0)
         cond_match->comp_value = b; /*holds pointer to b (!) for pattern subst*/
