@@ -46,10 +46,11 @@ struct data_config {
 };
 
 __attribute_cold__
+__attribute_returns_nonnull__
 data_config *data_config_init(void);
 
 __attribute_cold__
-int data_config_pcre_compile(data_config *dc);
+int data_config_pcre_compile(data_config *dc, int pcre_jit, log_error_st *errh);
 /*struct cond_cache_t;*/    /* declaration */ /*(moved to plugin_config.h)*/
 /*int data_config_pcre_exec(const data_config *dc, struct cond_cache_t *cache, buffer *b);*/
 
@@ -77,4 +78,7 @@ int config_parse_file(server *srv, config_t *context, const char *fn);
 __attribute_cold__
 int config_parse_cmd(server *srv, config_t *context, const char *cmd);
 
+__attribute_cold__
+int config_remoteip_normalize(buffer *b, buffer *tb);
+
 #endif
