@@ -4,9 +4,8 @@
 #include "fdevent.h"
 #include "buffer.h"
 
-#include <sys/time.h>
+#include "sys-time.h"
 
-#include <unistd.h>
 #include <stdlib.h>
 #include <string.h>
 
