@@ -24,10 +24,12 @@ typedef struct {
 } pcre_keyvalue_buffer;
 
 __attribute_cold__
+__attribute_malloc__
+__attribute_returns_nonnull__
 pcre_keyvalue_buffer *pcre_keyvalue_buffer_init(void);
 
 __attribute_cold__
-int pcre_keyvalue_buffer_append(log_error_st *errh, pcre_keyvalue_buffer *kvb, const buffer *key, const buffer *value);
+int pcre_keyvalue_buffer_append(log_error_st *errh, pcre_keyvalue_buffer *kvb, const buffer *key, const buffer *value, int pcre_jit);
 
 __attribute_cold__
 void pcre_keyvalue_buffer_free(pcre_keyvalue_buffer *kvb);
