@@ -27,20 +27,19 @@
 #include <crypt.h>
 #endif
 
-#include <mysql.h>
-
-#include "base.h"
-#include "http_auth.h"
-#include "log.h"
-#include "plugin.h"
-#include "safe_memclear.h"
-
-#include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
+#include <mysql.h>
+
+#include "mod_auth_api.h"
 #include "sys-crypto-md.h"
 
+#include "base.h"
+#include "ck.h"
+#include "log.h"
+#include "plugin.h"
+
 typedef struct {
     int auth_mysql_port;
     const char *auth_mysql_host;
@@ -267,7 +266,7 @@ SETDEFAULTS_FUNC(mod_authn_mysql_set_defaults) {
               case 7: /* auth.backend.mysql.col_user */
               case 8: /* auth.backend.mysql.col_pass */
               case 9: /* auth.backend.mysql.col_realm */
-                if (buffer_string_is_empty(cpv->v.b)) {
+                if (buffer_is_blank(cpv->v.b)) {
                     log_error(srv->errh, __FILE__, __LINE__,
                       "%s must not be blank", cpk[cpv->k_id].k);
                     return HANDLER_ERROR;
@@ -300,25 +299,21 @@ static int mod_authn_mysql_password_cmp(const char *userpw, unsigned long userpw
         char *crypted = crypt(reqpw, userpw);
         size_t crypwlen = (NULL != crypted) ? strlen(crypted) : 0;
         int rc = (crypwlen == userpwlen) ? memcmp(crypted, userpw, crypwlen) : -1;
-        if (crypwlen) safe_memclear(crypted, crypwlen);
+        if (crypwlen) ck_memzero(crypted, crypwlen);
         return rc;
     }
     else
   #endif
     if (32 == userpwlen) {
         /* plain md5 */
-        li_MD5_CTX Md5Ctx;
-        unsigned char HA1[16];
-        unsigned char md5pw[16];
-
-        li_MD5_Init(&Md5Ctx);
-        li_MD5_Update(&Md5Ctx, (unsigned char *)reqpw, strlen(reqpw));
-        li_MD5_Final(HA1, &Md5Ctx);
+        unsigned char HA1[MD5_DIGEST_LENGTH];
+        unsigned char md5pw[MD5_DIGEST_LENGTH];
+        MD5_once(HA1, reqpw, strlen(reqpw));
 
         /*(compare 16-byte MD5 binary instead of converting to hex strings
          * in order to then have to do case-insensitive hex str comparison)*/
-        return (0 == http_auth_digest_hex2bin(userpw, 32, md5pw, sizeof(md5pw)))
-          ? http_auth_const_time_memeq(HA1, md5pw, sizeof(md5pw)) ? 0 : 1
+        return (0 == li_hex2bin(md5pw, sizeof(md5pw), userpw, 32))
+          ? ck_memeq_const_time_fixed_len(HA1, md5pw, sizeof(md5pw)) ? 0 : 1
           : -1;
     }
 
@@ -353,8 +348,8 @@ static int mod_authn_mysql_result(plugin_data *p, http_auth_info_t *ai, const ch
         else {          /* used with HTTP Digest auth */
             /*(currently supports only single row, single digest algorithm)*/
             if (lengths[0] == (ai->dlen << 1)) {
-                rc = http_auth_digest_hex2bin(row[0], lengths[0],
-                                              ai->digest, sizeof(ai->digest));
+                rc = li_hex2bin(ai->digest, sizeof(ai->digest),
+                                row[0], lengths[0]);
             }
         }
     }
@@ -385,7 +380,7 @@ static handler_t mod_authn_mysql_query(request_st * const r, void *p_d, http_aut
     }
 
     do {
-        char q[1024], uname[512], urealm[512];
+        char uname[512], urealm[512];
         unsigned long mrc;
 
         if (ai->ulen > sizeof(uname)/2-1)
@@ -415,19 +410,25 @@ static handler_t mod_authn_mysql_query(request_st * const r, void *p_d, http_aut
         if ((unsigned long)~0 == mrc) break;
       #endif
 
-        rc = snprintf(q, sizeof(q),
-                      "SELECT %s FROM %s WHERE %s='%s' AND %s='%s'",
-                      p->conf.auth_mysql_col_pass,
-                      p->conf.auth_mysql_users_table,
-                      p->conf.auth_mysql_col_user,
-                      uname,
-                      p->conf.auth_mysql_col_realm,
-                      urealm);
-
-        if (rc >= (int)sizeof(q)) {
-            rc = -1;
-            break;
-        }
+        buffer * const tb = r->tmp_buf;
+        buffer_clear(tb);
+        struct const_iovec iov[] = {
+          { CONST_STR_LEN("SELECT ") }
+         ,{ p->conf.auth_mysql_col_pass, strlen(p->conf.auth_mysql_col_pass) }
+         ,{ CONST_STR_LEN(" FROM ") }
+         ,{ p->conf.auth_mysql_users_table, strlen(p->conf.auth_mysql_users_table) }
+         ,{ CONST_STR_LEN(" WHERE ") }
+         ,{ p->conf.auth_mysql_col_user, strlen(p->conf.auth_mysql_col_user) }
+         ,{ CONST_STR_LEN("='") }
+         ,{ uname, strlen(uname) }
+         ,{ CONST_STR_LEN("' AND ") }
+         ,{ p->conf.auth_mysql_col_realm, strlen(p->conf.auth_mysql_col_realm) }
+         ,{ CONST_STR_LEN("='") }
+         ,{ urealm, strlen(urealm) }
+         ,{ CONST_STR_LEN("'") }
+        };
+        buffer_append_iovec(tb, iov, sizeof(iov)/sizeof(*iov));
+        char * const q = tb->ptr;
 
         /* for now we stay synchronous */
         if (0 != mysql_query(p->mysql_conn, q)) {
@@ -467,9 +468,9 @@ static handler_t mod_authn_mysql_basic(request_st * const r, void *p_d, const ht
     ai.dalgo    = HTTP_AUTH_DIGEST_NONE;
     ai.dlen     = 0;
     ai.username = username->ptr;
-    ai.ulen     = buffer_string_length(username);
+    ai.ulen     = buffer_clen(username);
     ai.realm    = require->realm->ptr;
-    ai.rlen     = buffer_string_length(require->realm);
+    ai.rlen     = buffer_clen(require->realm);
     rc = mod_authn_mysql_query(r, p_d, &ai, pw);
     if (HANDLER_GO_ON != rc) return rc;
     return http_auth_match_rules(require, username->ptr, NULL, NULL)
