@@ -8,7 +8,6 @@
 #include "plugin.h"
 
 #include <sys/stat.h>
-#include <time.h>
 
 #include <stdlib.h>
 #include <string.h>
@@ -58,7 +57,7 @@ static int mod_cml_init_memcached(server *srv, config_plugin_value_t * const cpv
         buffer_append_string_buffer(opts, &ds->value);
     }
 
-    cpv->v.v = memcached(opts->ptr+1, buffer_string_length(opts)-1);
+    cpv->v.v = memcached(opts->ptr+1, buffer_clen(opts)-1);
 
     if (cpv->v.v) {
         cpv->vtype = T_CONFIG_LOCAL;
@@ -146,6 +145,8 @@ SETDEFAULTS_FUNC(mod_cml_set_defaults) {
         for (; -1 != cpv->k_id; ++cpv) {
             switch (cpv->k_id) {
               case 0: /* cml.extension */
+                if (buffer_is_blank(cpv->v.b))
+                    cpv->v.b = NULL;
                 break;
               case 1: /* cml.memcache-hosts */ /* config converted to memc handles */
                 if (!mod_cml_init_memcached(srv, cpv)) {
@@ -153,7 +154,10 @@ SETDEFAULTS_FUNC(mod_cml_set_defaults) {
                 }
                 break;
               case 2: /* cml.memcache-namespace *//*(unused)*/
+                break;
               case 3: /* cml.power-magnet */
+                if (buffer_is_blank(cpv->v.b))
+                    cpv->v.b = NULL;
                 break;
               default:/* should not happen */
                 break;
@@ -178,18 +182,18 @@ static int cache_call_lua(request_st * const r, plugin_data * const p, const buf
 	/* cleanup basedir */
 	b = &p->baseurl;
 	buffer_copy_buffer(b, &r->uri.path);
-	for (c = b->ptr + buffer_string_length(b); c > b->ptr && *c != '/'; c--);
+	for (c = b->ptr + buffer_clen(b); c > b->ptr && *c != '/'; c--);
 
 	if (*c == '/') {
-		buffer_string_set_length(b, c - b->ptr + 1);
+		buffer_truncate(b, c - b->ptr + 1);
 	}
 
 	b = &p->basedir;
 	buffer_copy_buffer(b, &r->physical.path);
-	for (c = b->ptr + buffer_string_length(b); c > b->ptr && *c != '/'; c--);
+	for (c = b->ptr + buffer_clen(b); c > b->ptr && *c != '/'; c--);
 
 	if (*c == '/') {
-		buffer_string_set_length(b, c - b->ptr + 1);
+		buffer_truncate(b, c - b->ptr + 1);
 	}
 
 
@@ -205,7 +209,7 @@ URIHANDLER_FUNC(mod_cml_power_magnet) {
 
 	mod_cml_patch_config(r, p);
 
-	if (buffer_string_is_empty(p->conf.power_magnet)) return HANDLER_GO_ON;
+	if (!p->conf.power_magnet) return HANDLER_GO_ON;
 
 	buffer_clear(&p->basedir);
 	buffer_clear(&p->baseurl);
@@ -241,7 +245,6 @@ URIHANDLER_FUNC(mod_cml_power_magnet) {
 			log_error(r->conf.errh, __FILE__, __LINE__, "cache-hit");
 		}
 		/* cache-hit */
-		buffer_reset(&r->physical.path);
 		return HANDLER_FINISHED;
 	case 1:
 		/* cache miss */
@@ -255,13 +258,15 @@ URIHANDLER_FUNC(mod_cml_power_magnet) {
 URIHANDLER_FUNC(mod_cml_is_handled) {
 	plugin_data *p = p_d;
 
-	if (buffer_string_is_empty(&r->physical.path)) return HANDLER_ERROR;
+	/* r->physical.path is non-empty for handle_subrequest_start */
+	/*if (buffer_is_blank(&r->physical.path)) return HANDLER_ERROR;*/
 
 	mod_cml_patch_config(r, p);
+	if (!p->conf.ext) return HANDLER_GO_ON;
 
-	if (buffer_string_is_empty(p->conf.ext)) return HANDLER_GO_ON;
-
-	if (!buffer_is_equal_right_len(&r->physical.path, p->conf.ext, buffer_string_length(p->conf.ext))) {
+	const uint32_t elen = buffer_clen(p->conf.ext);
+	const uint32_t plen = buffer_clen(&r->physical.path);
+	if (plen < elen || 0 != memcmp(r->physical.path.ptr+plen-elen, p->conf.ext->ptr, elen)) {
 		return HANDLER_GO_ON;
 	}
 
@@ -282,7 +287,6 @@ URIHANDLER_FUNC(mod_cml_is_handled) {
 			log_error(r->conf.errh, __FILE__, __LINE__, "cache-hit");
 		}
 		/* cache-hit */
-		buffer_reset(&r->physical.path);
 		return HANDLER_FINISHED;
 	case 1:
 		if (r->conf.log_request_handling) {
