@@ -109,6 +109,7 @@ typedef struct {
     plugin_config defaults;
     plugin_config conf;
     array *default_headers;
+    array tokens;
 } plugin_data;
 
 static plugin_data *mod_extforward_plugin_data_singleton;
@@ -120,7 +121,7 @@ static int extforward_check_proxy;
 typedef struct {
 	/* per-request state */
 	sock_addr saved_remote_addr;
-	buffer *saved_remote_addr_buf;
+	buffer saved_remote_addr_buf;
 
 	/* hap-PROXY protocol prior to receiving first request */
 	int(*saved_network_read)(connection *, chunkqueue *, off_t);
@@ -140,6 +141,7 @@ static handler_ctx * handler_ctx_init(void) {
 }
 
 static void handler_ctx_free(handler_ctx *hctx) {
+	free(hctx->saved_remote_addr_buf.ptr);
 	free(hctx);
 }
 
@@ -150,6 +152,7 @@ INIT_FUNC(mod_extforward_init) {
 FREE_FUNC(mod_extforward_free) {
     plugin_data * const p = p_d;
     array_free(p->default_headers);
+    array_free_data(&p->tokens);
     if (NULL == p->cvlist) return;
     /* (init i to 0 if global context; to 1 to skip empty global context) */
     for (int i = !p->cvlist[0].v.u2[1], used = p->nconfig; i < used; ++i) {
@@ -250,7 +253,7 @@ static void * mod_extforward_parse_forwarder(server *srv, const array *forwarder
         data_string * const ds = (data_string *)forwarder->data[j];
         char * const nm_slash = strchr(ds->key.ptr, '/');
         if (NULL == nm_slash) continue;
-        if (buffer_string_is_empty(&ds->value)) continue; /* ignored */
+        if (buffer_is_blank(&ds->value)) continue; /* ignored */
 
         char *err;
         const int nm_bits = strtol(nm_slash + 1, &err, 10);
@@ -368,7 +371,7 @@ SETDEFAULTS_FUNC(mod_extforward_set_defaults) {
                     for (uint32_t j = 0; j < a->used; ++j) {
                         data_string * const ds = (data_string *)a->data[j];
                         ds->ext =
-                          http_header_hkey_get(CONST_BUF_LEN(&ds->value));
+                          http_header_hkey_get(BUF_PTR_LEN(&ds->value));
                     }
                 }
                 break;
@@ -407,7 +410,7 @@ SETDEFAULTS_FUNC(mod_extforward_set_defaults) {
         array_insert_value(p->default_headers,CONST_STR_LEN("Forwarded-For"));
         for (uint32_t i = 0; i < p->default_headers->used; ++i) {
             data_string * const ds = (data_string *)p->default_headers->data[i];
-            ds->ext = http_header_hkey_get(CONST_BUF_LEN(&ds->value));
+            ds->ext = http_header_hkey_get(BUF_PTR_LEN(&ds->value));
         }
     }
 
@@ -456,10 +459,9 @@ SETDEFAULTS_FUNC(mod_extforward_set_defaults) {
 /*
    extract a forward array from the environment
 */
-static array *extract_forward_array(const buffer *pbuffer)
+static void extract_forward_array(array * const result, const buffer *pbuffer)
 {
-	array *result = array_init(8);
-	if (!buffer_string_is_empty(pbuffer)) {
+		/*force_assert(!buffer_is_blank(pbuffer));*/
 		const char *base, *curr;
 		/* state variable, 0 means not in string, 1 means in string */
 		int in_str = 0;
@@ -484,8 +486,6 @@ static array *extract_forward_array(const buffer *pbuffer)
 		if (in_str) {
 			array_insert_value(result, base, curr - base);
 		}
-	}
-	return result;
 }
 
 /*
@@ -495,7 +495,7 @@ static int is_proxy_trusted(plugin_data *p, const char * const ip, size_t iplen)
 {
     const data_string *ds =
       (const data_string *)array_get_element_klen(p->conf.forwarder, ip, iplen);
-    if (NULL != ds) return !buffer_string_is_empty(&ds->value);
+    if (NULL != ds) return !buffer_is_blank(&ds->value);
 
     if (p->conf.forward_masks_used) {
         const struct sock_addr_mask * const addrs = p->conf.forward_masks;
@@ -522,38 +522,38 @@ static int is_proxy_trusted(plugin_data *p, const char * const ip, size_t iplen)
 static int is_connection_trusted(connection * const con, plugin_data *p)
 {
     if (p->conf.forward_all) return (1 == p->conf.forward_all);
-    return is_proxy_trusted(p, CONST_BUF_LEN(con->dst_addr_buf));
+    return is_proxy_trusted(p, BUF_PTR_LEN(&con->dst_addr_buf));
 }
 
 /*
  * Return last address of proxy that is not trusted.
  * Do not accept "all" keyword here.
  */
-static const char *last_not_in_array(array *a, plugin_data *p)
+static const buffer *last_not_in_array(array *a, plugin_data *p)
 {
 	int i;
 
 	for (i = a->used - 1; i >= 0; i--) {
 		data_string *ds = (data_string *)a->data[i];
-		if (!is_proxy_trusted(p, CONST_BUF_LEN(&ds->value))) {
-			return ds->value.ptr;
+		if (!is_proxy_trusted(p, BUF_PTR_LEN(&ds->value))) {
+			return &ds->value;
 		}
 	}
 	return NULL;
 }
 
-static int mod_extforward_set_addr(request_st * const r, plugin_data *p, const char *addr) {
+static int mod_extforward_set_addr(request_st * const r, plugin_data *p, const char *addr, size_t addrlen) {
 	connection * const con = r->con;
 	sock_addr sock;
 	handler_ctx *hctx = con->plugin_ctx[p->id];
 
 	/* Preserve changed addr for lifetime of h2 connection; upstream proxy
 	 * should not reuse same h2 connection for requests from different clients*/
-	if (hctx && NULL != hctx->saved_remote_addr_buf
+	if (hctx && !buffer_is_unset(&hctx->saved_remote_addr_buf)
 	    && r->http_version > HTTP_VERSION_1_1) { /*(e.g. HTTP_VERSION_2)*/
 		if (extforward_check_proxy) /* save old address */
 			http_header_env_set(r, CONST_STR_LEN("_L_EXTFORWARD_ACTUAL_FOR"),
-			                    CONST_BUF_LEN(hctx->saved_remote_addr_buf));
+			                    BUF_PTR_LEN(&hctx->saved_remote_addr_buf));
 		return 1;
 	}
 
@@ -567,34 +567,28 @@ static int mod_extforward_set_addr(request_st * const r, plugin_data *p, const c
 
 	/* we found the remote address, modify current connection and save the old address */
 	if (hctx) {
-		if (hctx->saved_remote_addr_buf) {
+		if (!buffer_is_unset(&hctx->saved_remote_addr_buf)) {
 			if (r->conf.log_request_handling) {
 				log_error(r->conf.errh, __FILE__, __LINE__,
 				  "-- mod_extforward_uri_handler already patched this connection, resetting state");
 			}
 			con->dst_addr = hctx->saved_remote_addr;
-			buffer_free(con->dst_addr_buf);
-			con->dst_addr_buf = hctx->saved_remote_addr_buf;
-			hctx->saved_remote_addr_buf = NULL;
+			buffer_move(&con->dst_addr_buf, &hctx->saved_remote_addr_buf);
 		}
 	} else {
 		con->plugin_ctx[p->id] = hctx = handler_ctx_init();
 	}
 	/* save old address */
 	if (extforward_check_proxy) {
-		http_header_env_set(r, CONST_STR_LEN("_L_EXTFORWARD_ACTUAL_FOR"), CONST_BUF_LEN(con->dst_addr_buf));
+		http_header_env_set(r, CONST_STR_LEN("_L_EXTFORWARD_ACTUAL_FOR"),
+		                    BUF_PTR_LEN(&con->dst_addr_buf));
 	}
 	hctx->request_count = con->request_count;
 	hctx->saved_remote_addr = con->dst_addr;
-	hctx->saved_remote_addr_buf = con->dst_addr_buf;
+	buffer_move(&hctx->saved_remote_addr_buf, &con->dst_addr_buf);
 	/* patch connection address */
 	con->dst_addr = sock;
-	con->dst_addr_buf = buffer_init_string(addr);
-
-	if (r->conf.log_request_handling) {
-		log_error(r->conf.errh, __FILE__, __LINE__,
-		  "patching con->dst_addr_buf for the accesslog: %s", addr);
-	}
+	buffer_copy_string_len(&con->dst_addr_buf, addr, addrlen);
 
 	/* Now, clean the conf_cond cache, because we may have changed the results of tests */
 	config_cond_cache_reset_item(r, COMP_HTTP_REMOTE_IP);
@@ -616,7 +610,7 @@ static void mod_extforward_set_proto(request_st * const r, const char * const pr
 		 *   (probably) or the original value.
 		 */
 		if (extforward_check_proxy) {
-			http_header_env_set(r, CONST_STR_LEN("_L_EXTFORWARD_ACTUAL_PROTO"), CONST_BUF_LEN(&r->uri.scheme));
+			http_header_env_set(r, CONST_STR_LEN("_L_EXTFORWARD_ACTUAL_PROTO"), BUF_PTR_LEN(&r->uri.scheme));
 		}
 		if (buffer_eq_icase_ss(proto, protolen, CONST_STR_LEN("https"))) {
 	                r->con->proto_default_port = 443; /* "https" */
@@ -632,8 +626,9 @@ static void mod_extforward_set_proto(request_st * const r, const char * const pr
 
 static handler_t mod_extforward_X_Forwarded_For(request_st * const r, plugin_data * const p, const buffer * const x_forwarded_for) {
 	/* build forward_array from forwarded data_string */
-	array *forward_array = extract_forward_array(x_forwarded_for);
-	const char *real_remote_addr = last_not_in_array(forward_array, p);
+	array * const forward_array = &p->tokens;
+	extract_forward_array(forward_array, x_forwarded_for);
+	const buffer *real_remote_addr = last_not_in_array(forward_array, p);
 	if (real_remote_addr != NULL) { /* parsed */
 		/* get scheme if X-Forwarded-Proto is set
 		 * Limitations:
@@ -644,14 +639,15 @@ static handler_t mod_extforward_X_Forwarded_For(request_st * const r, plugin_dat
 		 *    as in X-Forwarded-For to find proto set by last trusted proxy)
 		 */
 		const buffer *x_forwarded_proto = http_header_request_get(r, HTTP_HEADER_X_FORWARDED_PROTO, CONST_STR_LEN("X-Forwarded-Proto"));
-		if (mod_extforward_set_addr(r, p, real_remote_addr) && NULL != x_forwarded_proto) {
-			mod_extforward_set_proto(r, CONST_BUF_LEN(x_forwarded_proto));
+		if (mod_extforward_set_addr(r, p, BUF_PTR_LEN(real_remote_addr)) && NULL != x_forwarded_proto) {
+			mod_extforward_set_proto(r, BUF_PTR_LEN(x_forwarded_proto));
 		}
 	}
-	array_free(forward_array);
+	array_reset_data_strings(forward_array);
 	return HANDLER_GO_ON;
 }
 
+__attribute_pure__
 static int find_end_quoted_string (const char * const s, int i) {
     do {
         ++i;
@@ -659,6 +655,7 @@ static int find_end_quoted_string (const char * const s, int i) {
     return i;
 }
 
+__attribute_pure__
 static int find_next_semicolon_or_comma_or_eq (const char * const s, int i) {
     for (; s[i] != '=' && s[i] != ';' && s[i] != ',' && s[i] != '\0'; ++i) {
         if (s[i] == '"') {
@@ -669,6 +666,7 @@ static int find_next_semicolon_or_comma_or_eq (const char * const s, int i) {
     return i;
 }
 
+__attribute_pure__
 static int find_next_semicolon_or_comma (const char * const s, int i) {
     for (; s[i] != ';' && s[i] != ',' && s[i] != '\0'; ++i) {
         if (s[i] == '"') {
@@ -682,7 +680,7 @@ static int find_next_semicolon_or_comma (const char * const s, int i) {
 static int buffer_backslash_unescape (buffer * const b) {
     /* (future: might move to buffer.c) */
     size_t j = 0;
-    size_t len = buffer_string_length(b);
+    size_t len = buffer_clen(b);
     char *p = memchr(b->ptr, '\\', len);
 
     if (NULL == p) return 1; /*(nothing to do)*/
@@ -694,10 +692,19 @@ static int buffer_backslash_unescape (buffer * const b) {
         }
         p[j++] = p[i];
     }
-    buffer_string_set_length(b, (size_t)(p+j - b->ptr));
+    buffer_truncate(b, (size_t)(p+j - b->ptr));
     return 1;
 }
 
+__attribute_cold__
+static handler_t mod_extforward_bad_request (request_st * const r, const unsigned int line, const char * const msg)
+{
+    r->http_status = 400; /* Bad Request */
+    r->handler_module = NULL;
+    log_error(r->conf.errh, __FILE__, line, "%s", msg);
+    return HANDLER_FINISHED;
+}
+
 static handler_t mod_extforward_Forwarded (request_st * const r, plugin_data * const p, const buffer * const forwarded) {
     /* HTTP list need not consist of param=value tokens,
      * but this routine expect such for HTTP Forwarded header
@@ -716,7 +723,7 @@ static handler_t mod_extforward_Forwarded (request_st * const r, plugin_data * c
      */
     char * const s = forwarded->ptr;
     int i = 0, j = -1, v, vlen, k, klen;
-    int used = (int)buffer_string_length(forwarded);
+    int used = (int)buffer_clen(forwarded);
     int ofor = -1, oproto, ohost, oby, oremote_user;
     int offsets[256];/*(~50 params is more than reasonably expected to handle)*/
     while (i < used) {
@@ -734,11 +741,8 @@ static handler_t mod_extforward_Forwarded (request_st * const r, plugin_data * c
         i = find_next_semicolon_or_comma_or_eq(s, i);
         if (i < 0) {
             /*(reject IP spoofing if attacker sets improper quoted-string)*/
-            log_error(r->conf.errh, __FILE__, __LINE__,
+            return mod_extforward_bad_request(r, __LINE__,
               "invalid quoted-string in Forwarded header");
-            r->http_status = 400; /* Bad Request */
-            r->handler_module = NULL;
-            return HANDLER_FINISHED;
         }
         if (s[i] != '=') continue;
         klen = i - k;
@@ -746,11 +750,8 @@ static handler_t mod_extforward_Forwarded (request_st * const r, plugin_data * c
         i = find_next_semicolon_or_comma(s, i);
         if (i < 0) {
             /*(reject IP spoofing if attacker sets improper quoted-string)*/
-            log_error(r->conf.errh, __FILE__, __LINE__,
+            return mod_extforward_bad_request(r, __LINE__,
               "invalid quoted-string in Forwarded header");
-            r->http_status = 400; /* Bad Request */
-            r->handler_module = NULL;
-            return HANDLER_FINISHED;
         }
         vlen = i - v;              /* might be 0 */
 
@@ -769,11 +770,8 @@ static handler_t mod_extforward_Forwarded (request_st * const r, plugin_data * c
 
     if (j >= (int)(sizeof(offsets)/sizeof(int))-4) {
         /* error processing Forwarded; too many params; fail closed */
-        log_error(r->conf.errh, __FILE__, __LINE__,
+        return mod_extforward_bad_request(r, __LINE__,
           "Too many params in Forwarded header");
-        r->http_status = 400; /* Bad Request */
-        r->handler_module = NULL;
-        return HANDLER_FINISHED;
     }
 
     if (-1 == j) return HANDLER_GO_ON;  /* make no changes */
@@ -804,11 +802,8 @@ static handler_t mod_extforward_Forwarded (request_st * const r, plugin_data * c
                 ++v;
                 do { --vlen; } while (vlen > v && s[vlen] != ']');
                 if (v == vlen) {
-                    log_error(r->conf.errh, __FILE__, __LINE__,
+                    return mod_extforward_bad_request(r, __LINE__,
                       "Invalid IPv6 addr in Forwarded header");
-                    r->http_status = 400; /* Bad Request */
-                    r->handler_module = NULL;
-                    return HANDLER_FINISHED;
                 }
             }
             else if (s[v] != '_' && s[v] != '/' && s[v] != 'u') {
@@ -848,7 +843,7 @@ static handler_t mod_extforward_Forwarded (request_st * const r, plugin_data * c
         char c = *ipend;
         int rc;
         *ipend = '\0';
-        rc = mod_extforward_set_addr(r, p, s+offsets[ofor+2]);
+        rc = mod_extforward_set_addr(r, p, s+offsets[ofor+2], offsets[ofor+3]);
         *ipend = c;
         if (!rc) return HANDLER_GO_ON; /* invalid addr; make no changes */
     }
@@ -894,7 +889,7 @@ static handler_t mod_extforward_Forwarded (request_st * const r, plugin_data * c
             break;
         }
     }
-    i = ++j;
+    i = j+1;
 
     if (-1 != oproto) {
         /* remove trailing spaces/tabs, and double-quotes from proto
@@ -932,11 +927,16 @@ static handler_t mod_extforward_Forwarded (request_st * const r, plugin_data * c
             j += 4; /*(k, klen, v, vlen come in sets of 4)*/
         }
         if (-1 != ohost) {
-            if (extforward_check_proxy
-                && !buffer_string_is_empty(r->http_host)) {
-                http_header_env_set(r,
-                                    CONST_STR_LEN("_L_EXTFORWARD_ACTUAL_HOST"),
-                                    CONST_BUF_LEN(r->http_host));
+            if (r->http_host && !buffer_is_blank(r->http_host)) {
+                if (extforward_check_proxy)
+                    http_header_env_set(r,
+                      CONST_STR_LEN("_L_EXTFORWARD_ACTUAL_HOST"),
+                      BUF_PTR_LEN(r->http_host));
+            }
+            else {
+                r->http_host =
+                  http_header_request_set_ptr(r, HTTP_HEADER_HOST,
+                                              CONST_STR_LEN("Host"));
             }
             /* remove trailing spaces/tabs, and double-quotes from host */
             v = offsets[ohost+2];
@@ -944,28 +944,22 @@ static handler_t mod_extforward_Forwarded (request_st * const r, plugin_data * c
             while (vlen > v && (s[vlen-1] == ' ' || s[vlen-1] == '\t')) --vlen;
             if (vlen > v+1 && s[v] == '"' && s[vlen-1] == '"') {
                 ++v; --vlen;
-                buffer_copy_string_len(r->http_host, s+v, vlen-v);
+                buffer_copy_string_len_lc(r->http_host, s+v, vlen-v);
                 if (!buffer_backslash_unescape(r->http_host)) {
-                    log_error(r->conf.errh, __FILE__, __LINE__,
+                    return mod_extforward_bad_request(r, __LINE__,
                       "invalid host= value in Forwarded header");
-                    r->http_status = 400; /* Bad Request */
-                    r->handler_module = NULL;
-                    return HANDLER_FINISHED;
                 }
             }
             else {
-                buffer_copy_string_len(r->http_host, s+v, vlen-v);
+                buffer_copy_string_len_lc(r->http_host, s+v, vlen-v);
             }
 
             if (0 != http_request_host_policy(r->http_host,
                                               r->conf.http_parseopts,
                                               r->con->proto_default_port)) {
                 /*(reject invalid chars in Host)*/
-                log_error(r->conf.errh, __FILE__, __LINE__,
+                return mod_extforward_bad_request(r, __LINE__,
                   "invalid host= value in Forwarded header");
-                r->http_status = 400; /* Bad Request */
-                r->handler_module = NULL;
-                return HANDLER_FINISHED;
             }
 
             config_cond_cache_reset_item(r, COMP_HTTP_HOST);
@@ -996,11 +990,8 @@ static handler_t mod_extforward_Forwarded (request_st * const r, plugin_data * c
                 euser = http_header_env_get(r, CONST_STR_LEN("REMOTE_USER"));
                 force_assert(NULL != euser);
                 if (!buffer_backslash_unescape(euser)) {
-                    log_error(r->conf.errh, __FILE__, __LINE__,
+                    return mod_extforward_bad_request(r, __LINE__,
                       "invalid remote_user= value in Forwarded header");
-                    r->http_status = 400; /* Bad Request */
-                    r->handler_module = NULL;
-                    return HANDLER_FINISHED;
                 }
             }
             else {
@@ -1015,13 +1006,14 @@ static handler_t mod_extforward_Forwarded (request_st * const r, plugin_data * c
         && !light_btst(r->rqst_htags, HTTP_HEADER_X_FORWARDED_FOR)) {
         /* create X-Forwarded-For if not present
          * (and at least original connecting IP is a trusted proxy) */
-        buffer *xff = r->tmp_buf;
-        buffer_clear(xff);
+        buffer * const xff =
+          http_header_request_set_ptr(r, HTTP_HEADER_X_FORWARDED_FOR,
+                                      CONST_STR_LEN("X-Forwarded-For"));
         for (j = 0; j < used; ) {
             if (-1 == offsets[j]) { ++j; continue; }
             if (3 == offsets[j+1]
                 && buffer_eq_icase_ssn(s+offsets[j], "for", 3)) {
-                if (!buffer_string_is_empty(xff))
+                if (!buffer_is_blank(xff))
                     buffer_append_string_len(xff, CONST_STR_LEN(", "));
                 /* quoted-string, IPv6 brackets, and :port already removed */
                 v = offsets[j+2];
@@ -1029,14 +1021,14 @@ static handler_t mod_extforward_Forwarded (request_st * const r, plugin_data * c
                 buffer_append_string_len(xff, s+v, vlen);
                 if (s[v-1] != '=') { /*(must have been quoted-string)*/
                     char *x =
-                      memchr(xff->ptr+buffer_string_length(xff)-vlen,'\\',vlen);
+                      memchr(xff->ptr + buffer_clen(xff) - vlen, '\\', vlen);
                     if (NULL != x) { /* backslash unescape in-place */
                         for (v = 0; x[v]; ++x) {
                             if (x[v] == '\\' && x[++v] == '\0')
                                 break; /*(invalid trailing backslash)*/
                             *x = x[v];
                         }
-                        buffer_string_set_length(xff, x - xff->ptr);
+                        buffer_truncate(xff, x - xff->ptr);
                     }
                 }
                 /* skip to next group; take first "for=..." in group
@@ -1047,7 +1039,6 @@ static handler_t mod_extforward_Forwarded (request_st * const r, plugin_data * c
             }
             j += 4; /*(k, klen, v, vlen come in sets of 4)*/
         }
-        http_header_request_set(r, HTTP_HEADER_X_FORWARDED_FOR, CONST_STR_LEN("X-Forwarded-For"), CONST_BUF_LEN(xff));
     }
   #endif
 
@@ -1056,17 +1047,9 @@ static handler_t mod_extforward_Forwarded (request_st * const r, plugin_data * c
 
 URIHANDLER_FUNC(mod_extforward_uri_handler) {
 	plugin_data *p = p_d;
-	const buffer *forwarded = NULL;
-	int is_forwarded_header = 0;
-
 	mod_extforward_patch_config(r, p);
 	if (NULL == p->conf.forwarder) return HANDLER_GO_ON;
 
-	if (r->conf.log_request_handling) {
-		log_error(r->conf.errh, __FILE__, __LINE__,
-		  "-- mod_extforward_uri_handler called");
-	}
-
 	if (p->conf.hap_PROXY_ssl_client_verify) {
 		const data_string *ds;
 		handler_ctx *hctx = r->con->plugin_ctx[p->id];
@@ -1077,7 +1060,7 @@ URIHANDLER_FUNC(mod_extforward_uri_handler) {
 					    CONST_STR_LEN("SUCCESS"));
 			http_header_env_set(r,
 					    CONST_STR_LEN("REMOTE_USER"),
-					    CONST_BUF_LEN(&ds->value));
+					    BUF_PTR_LEN(&ds->value));
 			http_header_env_set(r,
 					    CONST_STR_LEN("AUTH_TYPE"),
 					    CONST_STR_LEN("SSL_CLIENT_VERIFY"));
@@ -1101,44 +1084,36 @@ URIHANDLER_FUNC(mod_extforward_uri_handler) {
 	 * that mod_magnet and mod_cgi with local-redir should not modify
 	 * Forwarded or related headers and expect effects here. */
 	handler_ctx *hctx = r->con->plugin_ctx[p->id];
-	if (NULL != hctx && NULL != hctx->saved_remote_addr_buf
+	if (NULL != hctx && !buffer_is_unset(&hctx->saved_remote_addr_buf)
 	    && hctx->request_count == r->con->request_count)
 		return HANDLER_GO_ON;
 
+	const buffer *forwarded = NULL;
+	int is_forwarded_header = 0;
 	for (uint32_t k = 0; k < p->conf.headers->used; ++k) {
 		const data_string * const ds = (data_string *)p->conf.headers->data[k];
 		const buffer * const hdr = &ds->value;
-		forwarded = http_header_request_get(r, ds->ext, CONST_BUF_LEN(hdr));
+		forwarded = http_header_request_get(r, ds->ext, BUF_PTR_LEN(hdr));
 		if (forwarded) {
 			is_forwarded_header = (ds->ext == HTTP_HEADER_FORWARDED);
 			break;
 		}
 	}
-	if (NULL == forwarded) {
-		if (r->conf.log_request_handling) {
-			log_error(r->conf.errh, __FILE__, __LINE__,
-			  "no forward header found, skipping");
-		}
 
-		return HANDLER_GO_ON;
+	if (forwarded && is_connection_trusted(r->con, p)) {
+		return (is_forwarded_header)
+		  ? mod_extforward_Forwarded(r, p, forwarded)
+		  : mod_extforward_X_Forwarded_For(r, p, forwarded);
 	}
-
-	/* if the remote ip itself is not trusted, then do nothing */
-	if (!is_connection_trusted(r->con, p)) {
+	else {
 		if (r->conf.log_request_handling) {
 			log_error(r->conf.errh, __FILE__, __LINE__,
+			  "no forward header found or "
 			  "remote address %s is NOT a trusted proxy, skipping",
-			  r->con->dst_addr_buf->ptr);
+			  r->con->dst_addr_buf.ptr);
 		}
-
 		return HANDLER_GO_ON;
 	}
-
-	if (is_forwarded_header) {
-		return mod_extforward_Forwarded(r, p, forwarded);
-	}
-
-	return mod_extforward_X_Forwarded_For(r, p, forwarded);
 }
 
 
@@ -1150,8 +1125,7 @@ REQUEST_FUNC(mod_extforward_handle_request_env) {
         /* note: replaces values which may have been set by mod_openssl
          * (when mod_extforward is listed after mod_openssl in server.modules)*/
         data_string *ds = (data_string *)hctx->env->data[i];
-        http_header_env_set(r,
-                            CONST_BUF_LEN(&ds->key), CONST_BUF_LEN(&ds->value));
+        http_header_env_set(r, BUF_PTR_LEN(&ds->key), BUF_PTR_LEN(&ds->value));
     }
     return HANDLER_GO_ON;
 }
@@ -1170,11 +1144,9 @@ REQUEST_FUNC(mod_extforward_restore) {
 
 	if (!hctx) return HANDLER_GO_ON;
 
-	if (NULL != hctx->saved_remote_addr_buf) {
+	if (!buffer_is_unset(&hctx->saved_remote_addr_buf)) {
 		con->dst_addr = hctx->saved_remote_addr;
-		buffer_free(con->dst_addr_buf);
-		con->dst_addr_buf = hctx->saved_remote_addr_buf;
-		hctx->saved_remote_addr_buf = NULL;
+		buffer_move(&con->dst_addr_buf, &hctx->saved_remote_addr_buf);
 		/* Now, clean the conf_cond cache, because we may have changed the results of tests */
 		config_cond_cache_reset_item(r, COMP_HTTP_REMOTE_IP);
 	}
@@ -1196,10 +1168,9 @@ CONNECTION_FUNC(mod_extforward_handle_con_close)
         if (NULL != hctx->saved_network_read) {
             con->network_read = hctx->saved_network_read;
         }
-        if (NULL != hctx->saved_remote_addr_buf) {
+        if (!buffer_is_unset(&hctx->saved_remote_addr_buf)) {
             con->dst_addr = hctx->saved_remote_addr;
-            buffer_free(con->dst_addr_buf);
-            con->dst_addr_buf = hctx->saved_remote_addr_buf;
+            buffer_move(&con->dst_addr_buf, &hctx->saved_remote_addr_buf);
         }
         if (NULL != hctx->env) {
             array_free(hctx->env);
@@ -1231,7 +1202,7 @@ CONNECTION_FUNC(mod_extforward_handle_con_accept)
         if (r->conf.log_request_handling) {
             log_error(r->conf.errh, __FILE__, __LINE__,
               "remote address %s is NOT a trusted proxy, skipping",
-              con->dst_addr_buf->ptr);
+              con->dst_addr_buf.ptr);
         }
     }
     return HANDLER_GO_ON;
@@ -1329,6 +1300,7 @@ The following types have already been registered for the <type> field :
 #define PP2_TYPE_AUTHORITY        0x02
 #define PP2_TYPE_CRC32C           0x03
 #define PP2_TYPE_NOOP             0x04
+#define PP2_TYPE_UNIQUE_ID        0x05
 #define PP2_TYPE_SSL              0x20
 #define PP2_SUBTYPE_SSL_VERSION   0x21
 #define PP2_SUBTYPE_SSL_CN        0x22
@@ -1515,7 +1487,7 @@ static int mod_extforward_hap_PROXY_v1 (connection * const con,
     /* re-parse addr to string to normalize
      * (instead of trusting PROXY to provide canonicalized src_addr string)
      * (should prefer PROXY v2 protocol if concerned about performance) */
-    sock_addr_inet_ntop_copy_buffer(con->dst_addr_buf, &con->dst_addr);
+    sock_addr_inet_ntop_copy_buffer(&con->dst_addr_buf, &con->dst_addr);
 
     return 0;
 }
@@ -1575,7 +1547,7 @@ static int mod_extforward_hap_PROXY_v2 (connection * const con,
       case 0x11:  /* TCPv4 */
         sock_addr_assign(&con->dst_addr, AF_INET, hdr->v2.addr.ip4.src_port,
                                                  &hdr->v2.addr.ip4.src_addr);
-        sock_addr_inet_ntop_copy_buffer(con->dst_addr_buf, &con->dst_addr);
+        sock_addr_inet_ntop_copy_buffer(&con->dst_addr_buf, &con->dst_addr);
        #if 0
         ((struct sockaddr_in *)&by)->sin_family = AF_INET;
         ((struct sockaddr_in *)&by)->sin_addr.s_addr =
@@ -1589,7 +1561,7 @@ static int mod_extforward_hap_PROXY_v2 (connection * const con,
       case 0x21:  /* TCPv6 */
         sock_addr_assign(&con->dst_addr, AF_INET6, hdr->v2.addr.ip6.src_port,
                                                   &hdr->v2.addr.ip6.src_addr);
-        sock_addr_inet_ntop_copy_buffer(con->dst_addr_buf, &con->dst_addr);
+        sock_addr_inet_ntop_copy_buffer(&con->dst_addr_buf, &con->dst_addr);
        #if 0
         ((struct sockaddr_in6 *)&by)->sin6_family = AF_INET6;
         memcpy(&((struct sockaddr_in6 *)&by)->sin6_addr,
@@ -1608,7 +1580,7 @@ static int mod_extforward_hap_PROXY_v2 (connection * const con,
             if (NULL == z) return -1; /* invalid addr; too long */
             len = (uint32_t)(z - src_addr + 1); /*(+1 for '\0')*/
             sock_addr_assign(&con->dst_addr, AF_UNIX, 0, src_addr);
-            buffer_copy_string_len(con->dst_addr_buf, src_addr, len);
+            buffer_copy_string_len(&con->dst_addr_buf, src_addr, len);
         }
        #if 0 /*(dst_addr should be identical to src_addr for AF_UNIX)*/
         ((struct sockaddr_un *)&by)->sun_family = AF_UNIX;
@@ -1624,11 +1596,17 @@ static int mod_extforward_hap_PROXY_v2 (connection * const con,
 
     /* (optional) Type-Length-Value (TLV vectors) follow addresses */
 
+    if (3 + len > sz) return 0;
+
+    handler_ctx * const hctx =
+      con->plugin_ctx[mod_extforward_plugin_data_singleton->id];
     tlv = (struct pp2_tlv *)((char *)hdr + 16);
     for (sz -= len, len -= 3; sz >= 3; sz -= 3 + len) {
         tlv = (struct pp2_tlv *)((char *)tlv + 3 + len);
         len = ((uint32_t)tlv->length_hi << 8) | tlv->length_lo;
         if (3 + len > sz) break; /*(invalid TLV)*/
+        const char *k;
+        uint32_t klen;
         switch (tlv->type) {
          #if 0 /*(not implemented here)*/
           case PP2_TYPE_ALPN:
@@ -1636,9 +1614,8 @@ static int mod_extforward_hap_PROXY_v2 (connection * const con,
           case PP2_TYPE_CRC32C:
          #endif
           case PP2_TYPE_SSL: {
+            if (len < 5) continue;
             static const uint32_t zero = 0;
-            handler_ctx *hctx =
-              con->plugin_ctx[mod_extforward_plugin_data_singleton->id];
             struct pp2_tlv_ssl *tlv_ssl =
               (struct pp2_tlv_ssl *)(void *)((char *)tlv+3);
             struct pp2_tlv *subtlv = tlv;
@@ -1649,53 +1626,56 @@ static int mod_extforward_hap_PROXY_v2 (connection * const con,
                 && 0 == memcmp(&tlv_ssl->verify, &zero, 4)) { /* misaligned */
                 hctx->ssl_client_verify = 1;
             }
+            if (len < 5 + 3) continue;
+            if (NULL == hctx->env) hctx->env = array_init(8);
             for (uint32_t subsz = len-5, n = 5; subsz >= 3; subsz -= 3 + n) {
                 subtlv = (struct pp2_tlv *)((char *)subtlv + 3 + n);
                 n = ((uint32_t)subtlv->length_hi << 8) | subtlv->length_lo;
                 if (3 + n > subsz) break; /*(invalid TLV)*/
-                if (NULL == hctx->env) hctx->env = array_init(8);
                 switch (subtlv->type) {
                   case PP2_SUBTYPE_SSL_VERSION:
-                    array_set_key_value(hctx->env,
-                                        CONST_STR_LEN("SSL_PROTOCOL"),
-                                        (char *)subtlv+3, n);
+                    k = "SSL_PROTOCOL";
+                    klen = sizeof("SSL_PROTOCOL")-1;
                     break;
                   case PP2_SUBTYPE_SSL_CN:
                     /* (tlv_ssl->client & PP2_CLIENT_CERT_CONN)
                      *   or
                      * (tlv_ssl->client & PP2_CLIENT_CERT_SESS) */
-                    array_set_key_value(hctx->env,
-                                        CONST_STR_LEN("SSL_CLIENT_S_DN_CN"),
-                                        (char *)subtlv+3, n);
+                    k = "SSL_CLIENT_S_DN_CN";
+                    klen = sizeof("SSL_CLIENT_S_DN_CN")-1;
                     break;
                   case PP2_SUBTYPE_SSL_CIPHER:
-                    array_set_key_value(hctx->env,
-                                        CONST_STR_LEN("SSL_CIPHER"),
-                                        (char *)subtlv+3, n);
+                    k = "SSL_CIPHER";
+                    klen = sizeof("SSL_CIPHER")-1;
                     break;
                   case PP2_SUBTYPE_SSL_SIG_ALG:
-                    array_set_key_value(hctx->env,
-                                        CONST_STR_LEN("SSL_SERVER_A_SIG"),
-                                        (char *)subtlv+3, n);
+                    k = "SSL_SERVER_A_SIG";
+                    klen = sizeof("SSL_SERVER_A_SIG")-1;
                     break;
                   case PP2_SUBTYPE_SSL_KEY_ALG:
-                    array_set_key_value(hctx->env,
-                                        CONST_STR_LEN("SSL_SERVER_A_KEY"),
-                                        (char *)subtlv+3, n);
+                    k = "SSL_SERVER_A_KEY";
+                    klen = sizeof("SSL_SERVER_A_KEY")-1;
                     break;
                   default:
-                    break;
+                    continue;
                 }
+                array_set_key_value(hctx->env, k, klen, (char *)subtlv+3, n);
             }
-            break;
+            continue;
           }
+          case PP2_TYPE_UNIQUE_ID:
+            k = "PP2_UNIQUE_ID";
+            klen = sizeof("PP2_UNIQUE_ID")-1;
+            break;
          #if 0 /*(not implemented here)*/
           case PP2_TYPE_NETNS:
          #endif
           /*case PP2_TYPE_NOOP:*//* no-op */
           default:
-            break;
+            continue;
         }
+        if (NULL == hctx->env) hctx->env = array_init(8);
+        array_set_key_value(hctx->env, k, klen, (char *)tlv+3, len);
     }
 
     return 0;
