@@ -95,7 +95,7 @@ FREE_FUNC(mod_geoip_free) {
 
 static int mod_geoip_open_db(server *srv, config_plugin_value_t * const cpv, int mem_cache) {
     /* country db filename is required! */
-    if (buffer_is_empty(cpv->v.b)) {
+    if (buffer_is_blank(cpv->v.b)) {
         cpv->v.v = NULL;
         return 1;
     }
@@ -208,7 +208,7 @@ SETDEFAULTS_FUNC(mod_geoip_set_defaults) {
 
 static handler_t mod_geoip_query (request_st * const r, plugin_data * const p) {
     GeoIPRecord *gir;
-    const char *remote_ip = r->con->dst_addr_buf->ptr;
+    const char *remote_ip = r->con->dst_addr_buf.ptr;
 
     if (NULL != http_header_env_get(r, CONST_STR_LEN("GEOIP_COUNTRY_CODE"))) {
         return HANDLER_GO_ON;
@@ -243,6 +243,9 @@ static handler_t mod_geoip_query (request_st * const r, plugin_data * const p) {
         http_header_env_set(r, CONST_STR_LEN("GEOIP_CITY_NAME"), gir->city, strlen(gir->city));
         http_header_env_set(r, CONST_STR_LEN("GEOIP_CITY_POSTAL_CODE"), gir->postal_code, strlen(gir->postal_code));
 
+        buffer_append_int(http_header_env_set_ptr(r, CONST_STR_LEN("GEOIP_CITY_DMA_CODE")), (intmax_t)gir->dma_code);
+        buffer_append_int(http_header_env_set_ptr(r, CONST_STR_LEN("GEOIP_CITY_AREA_CODE")), (intmax_t)gir->area_code);
+
         {
             char latitude[32];
             snprintf(latitude, sizeof(latitude), "%f", gir->latitude);
@@ -255,18 +258,6 @@ static handler_t mod_geoip_query (request_st * const r, plugin_data * const p) {
             http_header_env_set(r, CONST_STR_LEN("GEOIP_CITY_LONG_LATITUDE"), long_latitude, strlen(long_latitude));
         }
 
-        {
-            char dc[LI_ITOSTRING_LENGTH];
-            http_header_env_set(r, CONST_STR_LEN("GEOIP_CITY_DMA_CODE"),
-                                dc, li_utostrn(dc, sizeof(dc), gir->dma_code));
-        }
-
-        {
-            char ac[LI_ITOSTRING_LENGTH];
-            http_header_env_set(r, CONST_STR_LEN("GEOIP_CITY_AREA_CODE"),
-                                ac, li_utostrn(ac, sizeof(ac), gir->area_code));
-        }
-
         GeoIPRecord_delete(gir);
     }
 
