@@ -8,12 +8,10 @@
 #include <lua.h>
 
 typedef struct {
-	buffer *name;
-	buffer *etag;
+	buffer name;
+	buffer etag;
 
 	lua_State *L;
-
-	time_t last_used; /* LRU */
 } script;
 
 typedef struct {
@@ -22,9 +20,22 @@ typedef struct {
 	uint32_t size;
 } script_cache;
 
+#if 0
+__attribute_cold__
+__attribute_malloc__
+__attribute_returns_nonnull__
 script_cache *script_cache_init(void);
+#endif
+
+__attribute_cold__
 void script_cache_free_data(script_cache *cache);
 
-lua_State *script_cache_get_script(script_cache *cache, buffer *name, int etag_flags);
+__attribute_cold__
+__attribute_nonnull__
+__attribute_returns_nonnull__
+script *script_cache_get_script(script_cache *cache, const buffer *name);
+
+__attribute_nonnull__
+lua_State *script_cache_check_script(script * const sc, int etag_flags);
 
 #endif
