@@ -7,9 +7,8 @@
 #include "rand.h"
 #include "chunk.h"
 #include "h2.h"             /* h2_send_1xx() */
-#include "http_auth.h"      /* http_auth_dumbdata_reset() */
-#include "http_vhostdb.h"   /* http_vhostdb_dumbdata_reset() */
 #include "fdevent.h"
+#include "fdlog.h"
 #include "connections.h"
 #include "sock_addr.h"
 #include "stat_cache.h"
@@ -29,15 +28,14 @@
 static const buffer default_server_tag = { CONST_STR_LEN(PACKAGE_DESC)+1, 0 };
 
 #include <sys/types.h>
-#include <sys/time.h>
 #include <sys/stat.h>
+#include "sys-time.h"
 
 #include <string.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <unistd.h>
 #include <stdlib.h>
-#include <time.h>
 #include <signal.h>
 #include <locale.h>
 
@@ -68,6 +66,19 @@ static const buffer default_server_tag = { CONST_STR_LEN(PACKAGE_DESC)+1, 0 };
 # include <sys/prctl.h>
 #endif
 
+#ifdef HAVE_MALLOC_H
+#ifndef LIGHTTPD_STATIC
+#ifdef HAVE_DLFCN_H
+#include <dlfcn.h>
+#endif
+#endif
+#include <malloc.h>
+#if defined(HAVE_MALLOC_TRIM)
+static int(*malloc_trim_fn)(size_t);
+static size_t malloc_top_pad;
+#endif
+#endif
+
 #include "sys-crypto.h"
 #if defined(USE_OPENSSL_CRYPTO) \
  || defined(USE_MBEDTLS_CRYPTO) \
@@ -97,7 +108,7 @@ static volatile sig_atomic_t srv_shutdown = 0;
 static volatile sig_atomic_t handle_sig_child = 0;
 static volatile sig_atomic_t handle_sig_alarm = 1;
 static volatile sig_atomic_t handle_sig_hup = 0;
-static time_t idle_limit = 0;
+static int idle_limit = 0;
 
 #if defined(HAVE_SIGACTION) && defined(SA_SIGINFO)
 static volatile siginfo_t last_sigterm_info;
@@ -186,7 +197,7 @@ static int daemonize(void) {
 	signal(SIGTSTP, SIG_IGN);
 #endif
 
-	if (pipe(pipefd) < 0) exit(-1);
+	if (fdevent_pipe_cloexec(pipefd, 64) < 0) exit(-1);
 
 	if (0 > (pid = fork())) exit(-1);
 
@@ -220,29 +231,80 @@ static int daemonize(void) {
 
 	if (0 != chdir("/")) exit(0);
 
-	fdevent_setfd_cloexec(pipefd[1]);
 	return pipefd[1];
 }
 #endif
 
+static int clockid_mono_coarse = 0;
+
+static unix_time64_t
+server_monotonic_secs (void)
+{
+    unix_timespec64_t ts;
+    return (0 == log_clock_gettime(clockid_mono_coarse, &ts))
+      ? ts.tv_sec
+      : log_monotonic_secs;
+}
+
+static unix_time64_t
+server_epoch_secs (server * const srv)
+{
+    const unix_time64_t cur_ts = log_epoch_secs;
+    const unix_time64_t new_ts = TIME64_CAST(time(NULL));
+    /* attempt to detect large clock jump */
+    if (new_ts < cur_ts || new_ts - cur_ts > 300) { /*(5 mins)*/
+        log_error(srv->errh, __FILE__, __LINE__,
+          "warning: clock jumped %lld secs",
+          (long long)((int64_t)new_ts - (int64_t)cur_ts));
+        int delta =                             /*(30 mins default)*/
+          config_feature_int(srv, "server.clock-jump-restart", 1800);
+        if (delta && (new_ts > cur_ts ? new_ts-cur_ts : cur_ts-new_ts) > delta){
+            log_error(srv->errh, __FILE__, __LINE__,
+              "attempting graceful restart in < ~5 seconds, else hard restart");
+            srv->graceful_expire_ts = log_monotonic_secs + 5;
+            raise(SIGUSR1);
+        }
+    }
+    return new_ts;
+}
+
 __attribute_cold__
+__attribute_noinline__
+__attribute_returns_nonnull__
 static server *server_init(void) {
 	server *srv = calloc(1, sizeof(*srv));
 	force_assert(srv);
-#define CLEAN(x) \
-	srv->x = buffer_init();
 
-	CLEAN(tmp_buf);
-#undef CLEAN
-	connection_joblist = &srv->joblist_A;
+	srv->tmp_buf = buffer_init();
 
 	strftime_cache_reset();
 
 	li_rand_reseed();
 
-	srv->startup_ts = log_epoch_secs = time(NULL);
+	srv->startup_ts = log_epoch_secs = TIME64_CAST(time(NULL));
+  #ifdef HAVE_CLOCK_GETTIME
+	unix_timespec64_t ts;
+	UNUSED(&ts);
+   #ifdef CLOCK_MONOTONIC_COARSE
+	if (0 == log_clock_gettime(CLOCK_MONOTONIC_COARSE, &ts))
+		clockid_mono_coarse = CLOCK_MONOTONIC_COARSE;
+	else
+   #endif
+   #ifdef CLOCK_MONOTONIC_RAW_APPROX
+	if (0 == log_clock_gettime(CLOCK_MONOTONIC_RAW_APPROX, &ts))
+		clockid_mono_coarse = CLOCK_MONOTONIC_RAW_APPROX;
+	else
+   #endif
+   #ifdef CLOCK_MONOTONIC_RAW
+	if (0 == log_clock_gettime(CLOCK_MONOTONIC_RAW, &ts))
+		clockid_mono_coarse = CLOCK_MONOTONIC_RAW;
+	else
+   #endif
+		clockid_mono_coarse = CLOCK_MONOTONIC;
+  #endif
+	log_monotonic_secs = server_monotonic_secs();
 
-	srv->errh = log_error_st_init();
+	srv->errh = log_set_global_errh(NULL, 0);
 
 	config_init(srv);
 
@@ -253,10 +315,13 @@ static server *server_init(void) {
 	srv->loadavg[2] = 0.0;
 	srv->stdin_fd = -1;
 
+	log_con_jqueue = (connection *)(uintptr_t)&log_con_jqueue;/*(sentinel)*/
+
 	return srv;
 }
 
 __attribute_cold__
+__attribute_noinline__
 static void server_free(server *srv) {
 	if (oneshot_fd > 0) {
 		if (oneshot_fdn) {
@@ -273,34 +338,27 @@ static void server_free(server *srv) {
 		close(srv->stdin_fd);
 	}
 
-#define CLEAN(x) \
-	buffer_free(srv->x);
-
-	CLEAN(tmp_buf);
-
-#undef CLEAN
+	buffer_free(srv->tmp_buf);
 
 	fdevent_free(srv->ev);
 
 	config_free(srv);
 
-	free(srv->joblist_A.ptr);
-	free(srv->joblist_B.ptr);
-	free(srv->fdwaitqueue.ptr);
-
 	stat_cache_free();
 
 	li_rand_cleanup();
 	chunkqueue_chunk_pool_free();
 
-	log_error_st_free(srv->errh);
+	if (srv->errh != log_set_global_errh(NULL, 0))
+		fdlog_free(srv->errh);
 	free(srv);
 }
 
 __attribute_cold__
+__attribute_noinline__
 static void remove_pid_file(server *srv) {
 	if (pid_fd <= -2) return;
-	if (!buffer_string_is_empty(srv->srvconf.pid_file) && 0 <= pid_fd) {
+	if (srv->srvconf.pid_file && 0 <= pid_fd) {
 		if (0 != ftruncate(pid_fd, 0)) {
 			log_perror(srv->errh, __FILE__, __LINE__,
 			  "ftruncate failed for: %s", srv->srvconf.pid_file->ptr);
@@ -310,8 +368,7 @@ static void remove_pid_file(server *srv) {
 		close(pid_fd);
 		pid_fd = -1;
 	}
-	if (!buffer_string_is_empty(srv->srvconf.pid_file) &&
-	    buffer_string_is_empty(srv->srvconf.changeroot)) {
+	if (srv->srvconf.pid_file && !srv->srvconf.changeroot) {
 		if (0 != unlink(srv->srvconf.pid_file->ptr)) {
 			if (errno != EACCES && errno != EPERM) {
 				log_perror(srv->errh, __FILE__, __LINE__,
@@ -417,7 +474,7 @@ static int server_oneshot_init_pipe(server *srv, int fdin, int fdout) {
      *   NCAT_PROTO (TCP, UDP, SCTP)
      */
     connection *con;
-    server_socket *srv_socket;
+    const server_socket *srv_socket;
     sock_addr cnt_addr;
 
     /* detect if called from netcat or else fabricate localhost addrs */
@@ -490,7 +547,7 @@ static int server_oneshot_init_pipe(server *srv, int fdin, int fdout) {
 __attribute_cold__
 static int server_oneshot_init(server *srv, int fd) {
 	connection *con;
-	server_socket *srv_socket;
+	const server_socket *srv_socket;
 	sock_addr cnt_addr;
 	socklen_t cnt_len;
 
@@ -672,7 +729,13 @@ static void show_features (void) {
 #endif
       ;
   show_version();
-  printf("%s%s%s\n", fdevent_show_event_handlers(), network_write_show_handlers(), features);
+  printf("%s%s%s%s\n",
+         fdevent_show_event_handlers(),
+         network_write_show_handlers(),
+         features,
+         sizeof(time_t) > 4 || (sizeof(time_t) == 4 && (time_t)-1 > (time_t)1)
+           ? "\t+ Y2038 support\n"
+           : "\t- Y2038 support (unsafe 32-bit signed time_t)\n");
 }
 
 __attribute_cold__
@@ -700,6 +763,7 @@ static void show_help (void) {
 }
 
 __attribute_cold__
+__attribute_noinline__
 static void server_sockets_save (server *srv) {    /* graceful_restart */
     for (uint32_t i = 0; i < srv->srv_sockets.used; ++i)
         srv->srv_sockets.ptr[i]->srv = NULL; /* srv will shortly be invalid */
@@ -712,6 +776,7 @@ static void server_sockets_save (server *srv) {    /* graceful_restart */
 }
 
 __attribute_cold__
+__attribute_noinline__
 static void server_sockets_restore (server *srv) { /* graceful_restart */
     memcpy(&srv->srv_sockets, &graceful_sockets, sizeof(server_socket_array));
     memset(&graceful_sockets, 0, sizeof(server_socket_array));
@@ -788,16 +853,15 @@ static int server_graceful_state_bg (server *srv) {
     /*assert(graceful_restart);*/
     /*(SIGUSR1 set to SIG_IGN in workers, so should not reach here if worker)*/
     if (srv_shutdown) return 0;
-    if (NULL == srv->srvconf.feature_flags) return 0;
 
     /* check if server should fork and background (bg) itself
      * to continue processing requests already in progress */
+    if (!config_feature_bool(srv, "server.graceful-restart-bg", 0)) return 0;
+
+    /*(set flag to false to avoid repeating)*/
     data_unset * const du =
       array_get_data_unset(srv->srvconf.feature_flags,
                            CONST_STR_LEN("server.graceful-restart-bg"));
-    if (!config_plugin_value_tobool(du, 0)) return 0;
-
-    /*(set flag to false to avoid repeating)*/
     if (du->type == TYPE_STRING)
         buffer_copy_string_len(&((data_string *)du)->value,
                                CONST_STR_LEN("false"));
@@ -828,7 +892,7 @@ static int server_graceful_state_bg (server *srv) {
     /* flush log buffers to avoid potential duplication of entries
      * server_handle_sighup(srv) does the following, but skip logging */
     plugins_call_handle_sighup(srv);
-    config_log_error_cycle(srv);
+    fdlog_files_cycle(srv->errh); /* reopen log files, not pipes */
 
     /* backgrounding to continue processing requests in progress */
     /* re-exec lighttpd in original process
@@ -904,7 +968,7 @@ static int server_graceful_state_bg (server *srv) {
      *  to allow graceful shutdown tasks to be run by server and by modules) */
     log_error(srv->errh, __FILE__, __LINE__,
       "[note] pid %lld continuing to handle %u connection(s) in progress",
-      (long long)getpid(), srv->conns.used);
+      (long long)getpid(), srv->srvconf.max_conns - srv->lim_conns);
 
     if (0 == srv->srvconf.max_worker) {
         /* reset graceful_shutdown; wait for signal from restarted server */
@@ -921,22 +985,21 @@ static void server_graceful_shutdown_maint (server *srv) {
     if (oneshot_fd) {
         /* permit keep-alive on one-shot connections until graceful_expire_ts */
         if (!srv->graceful_expire_ts) return;
-        if (srv->graceful_expire_ts >= log_epoch_secs) return;
+        if (srv->graceful_expire_ts >= log_monotonic_secs) return;
     }
     connection_graceful_shutdown_maint(srv);
 }
 
 __attribute_cold__
+__attribute_noinline__
 static void server_graceful_state (server *srv) {
 
     if (!srv_shutdown) {
-        if (0 == srv->graceful_expire_ts && srv->srvconf.feature_flags) {
-            const data_unset * const du =
-              array_get_element_klen(srv->srvconf.feature_flags,
-                CONST_STR_LEN("server.graceful-shutdown-timeout"));
-            srv->graceful_expire_ts = config_plugin_value_to_int32(du, 0);
+        if (0 == srv->graceful_expire_ts) {
+            srv->graceful_expire_ts =
+              config_feature_int(srv, "server.graceful-shutdown-timeout", 0);
             if (srv->graceful_expire_ts)
-                srv->graceful_expire_ts += log_epoch_secs;
+                srv->graceful_expire_ts += log_monotonic_secs;
         }
         server_graceful_shutdown_maint(srv);
     }
@@ -949,8 +1012,7 @@ static void server_graceful_state (server *srv) {
     log_error(srv->errh,__FILE__,__LINE__,"[note] graceful shutdown started");
 
     /* no graceful restart if chroot()ed, if oneshot mode, or if idle timeout */
-    if (!buffer_string_is_empty(srv->srvconf.changeroot)
-        || oneshot_fd || 2 == graceful_shutdown)
+    if (srv->srvconf.changeroot || oneshot_fd || 2 == graceful_shutdown)
         graceful_restart = 0;
 
     if (graceful_restart) {
@@ -967,6 +1029,7 @@ static void server_graceful_state (server *srv) {
 }
 
 __attribute_cold__
+__attribute_noinline__
 static void server_sockets_enable (server *srv) {
     server_sockets_set_event(srv, FDEVENT_IN);
     srv->sockets_disabled = 0;
@@ -974,45 +1037,30 @@ static void server_sockets_enable (server *srv) {
 }
 
 __attribute_cold__
+__attribute_noinline__
 static void server_sockets_disable (server *srv) {
     server_sockets_set_event(srv, 0);
     srv->sockets_disabled = 1;
     log_error(srv->errh, __FILE__, __LINE__,
-      (srv->conns.used >= srv->max_conns)
+      (0 == srv->lim_conns)
         ? "[note] sockets disabled, connection limit reached"
         : "[note] sockets disabled, out-of-fds");
 }
 
 __attribute_cold__
 static void server_overload_check (server *srv) {
-    if (srv->cur_fds + (int)srv->fdwaitqueue.used < srv->max_fds_lowat
-        && srv->conns.used < srv->max_conns) {
-
+    if (srv->cur_fds < srv->max_fds_lowat && 0 != srv->lim_conns)
         server_sockets_enable(srv);
-    }
 }
 
 static void server_load_check (server *srv) {
     /* check if hit limits for num fds used or num connections */
-    if (srv->cur_fds > srv->max_fds_hiwat || srv->conns.used >= srv->max_conns)
+    if (srv->cur_fds > srv->max_fds_hiwat || 0 == srv->lim_conns)
         server_sockets_disable(srv);
 }
 
 __attribute_cold__
 __attribute_noinline__
-static void server_process_fdwaitqueue (server *srv) {
-    connections * const fdwaitqueue = &srv->fdwaitqueue;
-    uint32_t i = 0;
-    for (int n = srv->max_fds - srv->cur_fds - 16; n > 0; --n) {
-        if (i == fdwaitqueue->used) break;
-        connection_state_machine(fdwaitqueue->ptr[i++]);
-    }
-    if (i > 0 && 0 != (fdwaitqueue->used -= i)) {
-	memmove(fdwaitqueue->ptr, fdwaitqueue->ptr+i, fdwaitqueue->used * sizeof(*(fdwaitqueue->ptr)));
-    }
-}
-
-__attribute_cold__
 static int server_main_setup (server * const srv, int argc, char **argv) {
 	int print_config = 0;
 	int test_config = 0;
@@ -1043,8 +1091,6 @@ static int server_main_setup (server * const srv, int argc, char **argv) {
 	handle_sig_hup = 0;
 	idle_limit = 0;
 	chunkqueue_set_tempdirs_default_reset();
-	http_auth_dumbdata_reset();
-	http_vhostdb_dumbdata_reset();
 	/*graceful_restart = 0;*//*(reset below to avoid further daemonizing)*/
 	/*(intentionally preserved)*/
 	/*memset(graceful_sockets, 0, sizeof(graceful_sockets));*/
@@ -1075,7 +1121,7 @@ static int server_main_setup (server * const srv, int argc, char **argv) {
 				  "Invalid idle timeout value: %s", optarg);
 				return -1;
 			}
-			idle_limit = (time_t)timeout;
+			idle_limit = (int)timeout;
 			break;
 		}
 		case 'p': print_config = 1; break;
@@ -1113,7 +1159,7 @@ static int server_main_setup (server * const srv, int argc, char **argv) {
 
 	if (print_config) {
 		config_print(srv);
-		fprintf(stdout, "\n");
+		puts(srv->tmp_buf->ptr);
 	}
 
 	if (test_config) {
@@ -1213,20 +1259,12 @@ static int server_main_setup (server * const srv, int argc, char **argv) {
 	}
 
 	http_response_send_1xx_cb_set(NULL, HTTP_VERSION_2);
-	if (srv->srvconf.feature_flags
-	    && !config_plugin_value_tobool(
-	          array_get_element_klen(srv->srvconf.feature_flags,
-	            CONST_STR_LEN("server.h2-discard-backend-1xx")), 0))
-		http_response_send_1xx_cb_set(h2_send_1xx,
-		                              HTTP_VERSION_2);
+	if (!config_feature_bool(srv, "server.h2-discard-backend-1xx", 0))
+		http_response_send_1xx_cb_set(h2_send_1xx, HTTP_VERSION_2);
 
 	http_response_send_1xx_cb_set(NULL, HTTP_VERSION_1_1);
-	if (srv->srvconf.feature_flags
-	    && !config_plugin_value_tobool(
-	          array_get_element_klen(srv->srvconf.feature_flags,
-	            CONST_STR_LEN("server.h1-discard-backend-1xx")), 0))
-		http_response_send_1xx_cb_set(connection_send_1xx,
-		                              HTTP_VERSION_1_1);
+	if (!config_feature_bool(srv, "server.h1-discard-backend-1xx", 0))
+		http_response_send_1xx_cb_set(connection_send_1xx, HTTP_VERSION_1_1);
 
 	if (0 != config_set_defaults(srv)) {
 		log_error(srv->errh, __FILE__, __LINE__,
@@ -1262,7 +1300,7 @@ static int server_main_setup (server * const srv, int argc, char **argv) {
 
 	/* open pid file BEFORE chroot */
 	if (-2 == pid_fd) pid_fd = -1; /*(initial startup state)*/
-	if (-1 == pid_fd && !buffer_string_is_empty(srv->srvconf.pid_file)) {
+	if (-1 == pid_fd && srv->srvconf.pid_file) {
 		const char *pidfile = srv->srvconf.pid_file->ptr;
 		if (-1 == (pid_fd = fdevent_open_cloexec(pidfile, 0, O_WRONLY | O_CREAT | O_EXCL | O_TRUNC, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH))) {
 			struct stat st;
@@ -1346,7 +1384,7 @@ static int server_main_setup (server * const srv, int argc, char **argv) {
 		struct group *grp = NULL;
 		struct passwd *pwd = NULL;
 
-		if (!buffer_string_is_empty(srv->srvconf.groupname)) {
+		if (srv->srvconf.groupname) {
 			if (NULL == (grp = getgrnam(srv->srvconf.groupname->ptr))) {
 				log_error(srv->errh, __FILE__, __LINE__,
 				  "can't find groupname %s", srv->srvconf.groupname->ptr);
@@ -1354,7 +1392,7 @@ static int server_main_setup (server * const srv, int argc, char **argv) {
 			}
 		}
 
-		if (!buffer_string_is_empty(srv->srvconf.username)) {
+		if (srv->srvconf.username) {
 			if (NULL == (pwd = getpwnam(srv->srvconf.username->ptr))) {
 				log_error(srv->errh, __FILE__, __LINE__,
 				  "can't find username %s", srv->srvconf.username->ptr);
@@ -1395,13 +1433,13 @@ static int server_main_setup (server * const srv, int argc, char **argv) {
 				log_perror(srv->errh, __FILE__, __LINE__, "setgroups()");
 				return -1;
 			}
-			if (!buffer_string_is_empty(srv->srvconf.username)) {
+			if (srv->srvconf.username) {
 				initgroups(srv->srvconf.username->ptr, grp->gr_gid);
 			}
 		}
 #endif
 #ifdef HAVE_CHROOT
-		if (!buffer_string_is_empty(srv->srvconf.changeroot)) {
+		if (srv->srvconf.changeroot) {
 			tzset();
 
 			if (-1 == chroot(srv->srvconf.changeroot->ptr)) {
@@ -1492,7 +1530,7 @@ static int server_main_setup (server * const srv, int argc, char **argv) {
 		buffer_clear(tb);
 		buffer_append_int(tb, srv->pid);
 		buffer_append_string_len(tb, CONST_STR_LEN("\n"));
-		if (-1 == write_all(pid_fd, CONST_BUF_LEN(tb))) {
+		if (-1 == write_all(pid_fd, BUF_PTR_LEN(tb))) {
 			log_perror(srv->errh, __FILE__, __LINE__, "Couldn't write pid file");
 			close(pid_fd);
 			pid_fd = -1;
@@ -1576,13 +1614,14 @@ static int server_main_setup (server * const srv, int argc, char **argv) {
 			} else {
 				int status;
 
-				if (-1 != (pid = fdevent_waitpid(-1, &status, 0))) {
-					log_epoch_secs = time(NULL);
+				if (-1 != (pid = fdevent_waitpid_intr(-1, &status))) {
+					log_monotonic_secs = server_monotonic_secs();
+					log_epoch_secs = server_epoch_secs(srv);
 					if (plugins_call_handle_waitpid(srv, pid, status) != HANDLER_GO_ON) {
 						if (!timer) alarm((timer = 5));
 						continue;
 					}
-					switch (fdevent_reaped_logger_pipe(pid)) {
+					switch (fdlog_pipes_waitpid_cb(pid)) {
 					  default: break;
 					  case -1: if (!timer) alarm((timer = 5));
 						   __attribute_fallthrough__
@@ -1601,15 +1640,15 @@ static int server_main_setup (server * const srv, int argc, char **argv) {
 				} else {
 					switch (errno) {
 					case EINTR:
-						log_epoch_secs = time(NULL);
+						log_monotonic_secs = server_monotonic_secs();
+						log_epoch_secs = server_epoch_secs(srv);
 						/**
 						 * if we receive a SIGHUP we have to close our logs ourself as we don't 
 						 * have the mainloop who can help us here
 						 */
 						if (handle_sig_hup) {
 							handle_sig_hup = 0;
-
-							config_log_error_cycle(srv);
+							fdlog_files_cycle(srv->errh); /* reopen log files, not pipes */
 
 							/* forward SIGHUP to workers */
 							for (int n = 0; n < npids; ++n) {
@@ -1620,7 +1659,7 @@ static int server_main_setup (server * const srv, int argc, char **argv) {
 							handle_sig_alarm = 0;
 							timer = 0;
 							plugins_call_handle_trigger(srv);
-							fdevent_restart_logger_pipes(log_epoch_secs);
+							fdlog_pipes_restart(log_monotonic_secs);
 						}
 						break;
 					default:
@@ -1670,13 +1709,15 @@ static int server_main_setup (server * const srv, int argc, char **argv) {
 		}
 		if (srv->srvconf.pid_file) buffer_clear(srv->srvconf.pid_file);
 
-		fdevent_clr_logger_pipe_pids();
+		fdlog_pipes_abandon_pids();
 		srv->pid = getpid();
 		li_rand_reseed();
 	}
 #endif
 
 	srv->max_fds = (int)srv->srvconf.max_fds;
+        if (srv->max_fds < 32) /*(sanity check; not expected)*/
+            srv->max_fds = 32; /*(server load checks will fail if too low)*/
 	srv->ev = fdevent_init(srv->srvconf.event_handler, &srv->max_fds, &srv->cur_fds, srv->errh);
 	if (NULL == srv->ev) {
 		log_error(srv->errh, __FILE__, __LINE__, "fdevent_init failed");
@@ -1692,17 +1733,15 @@ static int server_main_setup (server * const srv, int argc, char **argv) {
 		log_error(srv->errh, __FILE__, __LINE__,
 		  "can't have more connections than fds/2: %hu %d",
 		  srv->srvconf.max_conns, srv->max_fds);
-		srv->max_conns = srv->max_fds/2;
+		srv->lim_conns = srv->srvconf.max_conns = srv->max_fds/2;
 	} else if (srv->srvconf.max_conns) {
 		/* otherwise respect the wishes of the user */
-		srv->max_conns = srv->srvconf.max_conns;
+		srv->lim_conns = srv->srvconf.max_conns;
 	} else {
 		/* or use the default: we really don't want to hit max-fds */
-		srv->max_conns = srv->max_fds/3;
+		srv->lim_conns = srv->srvconf.max_conns = srv->max_fds/3;
 	}
 
-	request_pool_init(srv->max_conns);
-
 	/* libev backend overwrites our SIGCHLD handler and calls waitpid on SIGCHLD; we want our own SIGCHLD handling. */
 #ifdef HAVE_SIGACTION
 	sigaction(SIGCHLD, &act, NULL);
@@ -1719,6 +1758,8 @@ static int server_main_setup (server * const srv, int argc, char **argv) {
 		return -1;
 	}
 
+	chunkqueue_internal_pipes(config_feature_bool(srv, "chunkqueue.splice", 1));
+
 	/* might fail if user is using fam (not gamin) and famd isn't running */
 	if (!stat_cache_init(srv->ev, srv->errh)) {
 		log_error(srv->errh, __FILE__, __LINE__,
@@ -1781,8 +1822,7 @@ static void server_handle_sighup (server * const srv) {
 			/* cycle logfiles */
 
 			plugins_call_handle_sighup(srv);
-
-			config_log_error_cycle(srv);
+			fdlog_files_cycle(srv->errh); /* reopen log files, not pipes */
 #ifdef HAVE_SIGACTION
 				log_error(srv->errh, __FILE__, __LINE__,
 				  "logfiles cycled UID = %d PID = %d",
@@ -1795,14 +1835,15 @@ static void server_handle_sighup (server * const srv) {
 }
 
 __attribute_noinline__
-static void server_handle_sigalrm (server * const srv, time_t min_ts, time_t last_active_ts) {
+static void server_handle_sigalrm (server * const srv, unix_time64_t mono_ts, unix_time64_t last_active_ts) {
 
 				plugins_call_handle_trigger(srv);
 
-				log_epoch_secs = min_ts;
+				log_monotonic_secs = mono_ts;
+				log_epoch_secs = server_epoch_secs(srv);
 
 				/* check idle time limit, if enabled */
-				if (idle_limit && idle_limit < min_ts - last_active_ts && !graceful_shutdown) {
+				if (idle_limit && idle_limit < mono_ts - last_active_ts && !graceful_shutdown) {
 					log_error(srv->errh, __FILE__, __LINE__,
 					  "[note] idle timeout %ds exceeded, "
 					  "initiating graceful shutdown", (int)idle_limit);
@@ -1816,19 +1857,27 @@ static void server_handle_sigalrm (server * const srv, time_t min_ts, time_t las
 
 			      #ifdef HAVE_GETLOADAVG
 				/* refresh loadavg data every 30 seconds */
-				if (srv->loadts + 30 < min_ts) {
+				if (srv->loadts + 30 < mono_ts) {
 					if (-1 != getloadavg(srv->loadavg, 3)) {
-						srv->loadts = min_ts;
+						srv->loadts = mono_ts;
 					}
 				}
 			      #endif
 
-				if (0 == (min_ts & 0x3f)) { /*(once every 64 secs)*/
+				if (0 == (mono_ts & 0x3f)) { /*(once every 64 secs)*/
+					/* free logger buffers every 64 secs */
+					fdlog_flushall(srv->errh);
 					/* free excess chunkqueue buffers every 64 secs */
 					chunkqueue_chunk_pool_clear();
+					/* clear request and connection pools every 64 secs */
+					request_pool_free();
+					connections_pool_clear(srv);
+				  #if defined(HAVE_MALLOC_TRIM)
+					if (malloc_trim_fn) malloc_trim_fn(malloc_top_pad);
+				  #endif
 					/* attempt to restart dead piped loggers every 64 secs */
 					if (0 == srv->srvconf.max_worker)
-						fdevent_restart_logger_pipes(min_ts);
+						fdlog_pipes_restart(mono_ts);
 				}
 				/* cleanup stat-cache */
 				stat_cache_trigger_cleanup();
@@ -1837,7 +1886,7 @@ static void server_handle_sigalrm (server * const srv, time_t min_ts, time_t las
 				/* if graceful_shutdown, accelerate cleanup of recently completed request/responses */
 				if (graceful_shutdown && !srv_shutdown)
 					server_graceful_shutdown_maint(srv);
-				connection_periodic_maint(srv, min_ts);
+				connection_periodic_maint(srv, mono_ts);
 }
 
 __attribute_noinline__
@@ -1852,7 +1901,7 @@ static void server_handle_sigchld (server * const srv) {
 					}
 					if (0 == srv->srvconf.max_worker) {
 						/* check piped-loggers and restart, even if shutting down */
-						if (fdevent_waitpid_logger_pipe_pid(pid, log_epoch_secs)) {
+						if (fdlog_pipes_waitpid_cb(pid)) {
 							continue;
 						}
 					}
@@ -1861,19 +1910,20 @@ static void server_handle_sigchld (server * const srv) {
 }
 
 __attribute_hot__
-static void server_run_con_queue (connections * const restrict joblist) {
-    connection * const * const restrict conlist = joblist->ptr;
-    const uint32_t used = joblist->used;
-    joblist->used = 0;
-    for (uint32_t i = 0; i < used; ++i) {
-        connection_state_machine(conlist[i]);
+__attribute_nonnull__
+static void server_run_con_queue (connection * const restrict joblist, const connection * const sentinel) {
+    for (connection *con = joblist, *jqnext; con != sentinel; con = jqnext) {
+        jqnext = con->jqnext;
+        con->jqnext = NULL;
+        connection_state_machine(con);
     }
 }
 
 __attribute_hot__
 __attribute_noinline__
 static void server_main_loop (server * const srv) {
-	time_t last_active_ts = time(NULL);
+	unix_time64_t last_active_ts = server_monotonic_secs();
+	log_epoch_secs = server_epoch_secs(srv);
 
 	while (!srv_shutdown) {
 
@@ -1887,9 +1937,9 @@ static void server_main_loop (server * const srv) {
 		if (handle_sig_alarm) {
 			handle_sig_alarm = 0;
 	      #endif
-			time_t min_ts = time(NULL);
-			if (min_ts != log_epoch_secs) {
-				server_handle_sigalrm(srv, min_ts, last_active_ts);
+			unix_time64_t mono_ts = server_monotonic_secs();
+			if (mono_ts != log_monotonic_secs) {
+				server_handle_sigalrm(srv, mono_ts, last_active_ts);
 			}
 	      #ifdef USE_ALARM
 		}
@@ -1902,7 +1952,7 @@ static void server_main_loop (server * const srv) {
 
 		if (graceful_shutdown) {
 			server_graceful_state(srv);
-			if (0 == srv->conns.used && graceful_shutdown) {
+			if (NULL == srv->conns && graceful_shutdown) {
 				/* we are in graceful shutdown phase and all connections are closed
 				 * we are ready to terminate without harming anyone */
 				srv_shutdown = 1;
@@ -1914,28 +1964,20 @@ static void server_main_loop (server * const srv) {
 			server_load_check(srv);
 		}
 
-		if (srv->fdwaitqueue.used) {
-			server_process_fdwaitqueue(srv);
-		}
+		static connection * const sentinel =
+		  (connection *)(uintptr_t)&log_con_jqueue;
+		connection * const joblist = log_con_jqueue;
+		log_con_jqueue = sentinel;
+		server_run_con_queue(joblist, sentinel);
 
-		connections * const joblist = connection_joblist;
-
-		if (fdevent_poll(srv->ev, joblist->used ? 0 : 1000) > 0) {
-			last_active_ts = log_epoch_secs;
-		}
-
-		connection_joblist = (joblist == &srv->joblist_A)
-		  ? &srv->joblist_B
-		  : &srv->joblist_A;
-
-		server_run_con_queue(joblist);
+		if (fdevent_poll(srv->ev, log_con_jqueue != sentinel ? 0 : 1000) > 0)
+			last_active_ts = log_monotonic_secs;
 	}
 }
 
 __attribute_cold__
-int main (int argc, char **argv) {
-    int rc;
-
+__attribute_noinline__
+static int main_init_once (void) {
   #ifdef HAVE_GETUID
   #ifndef HAVE_ISSETUGID
   #define issetugid() (geteuid() != getuid() || getegid() != getgid())
@@ -1943,14 +1985,52 @@ int main (int argc, char **argv) {
     if (0 != getuid() && issetugid()) { /*check as early as possible in main()*/
         fprintf(stderr,
                 "Are you nuts ? Don't apply a SUID bit to this binary\n");
-        return -1;
+        return 0;
     }
   #endif
 
+  #if defined(HAVE_MALLOPT) && defined(M_ARENA_MAX)
+  #ifdef LIGHTTPD_STATIC
+    mallopt(M_ARENA_MAX, 2); /*(ignore error, if any)*/
+  #else
+    {
+        int (*mallopt_fn)(int, int);
+        mallopt_fn = (int (*)(int, int))(intptr_t)dlsym(RTLD_DEFAULT,"mallopt");
+        if (mallopt_fn) mallopt_fn(M_ARENA_MAX, 2); /*(ignore error, if any)*/
+    }
+  #endif
+  #endif
+
+  #if defined(HAVE_MALLOC_TRIM)
+    malloc_top_pad = 524288;
+    {
+        const char * const top_pad_str = getenv("MALLOC_TOP_PAD_");
+        if (top_pad_str) {
+            unsigned long top_pad = strtoul(top_pad_str, NULL, 10);
+            if (top_pad != ULONG_MAX) malloc_top_pad = (size_t)top_pad;
+        }
+    }
+  #ifdef LIGHTTPD_STATIC
+    malloc_trim_fn = malloc_trim;
+  #else
+    malloc_trim_fn =
+      (int (*)(size_t))(intptr_t)dlsym(RTLD_DEFAULT,"malloc_trim");
+  #endif
+  #endif
+
     /* for nice %b handling in strftime() */
     setlocale(LC_TIME, "C");
     tzset();
 
+    return 1;
+}
+
+__attribute_cold__
+int main (int argc, char ** argv) {
+    if (!main_init_once()) return -1;
+
+    int rc;
+
     do {
         server * const srv = server_init();
 
@@ -1968,6 +2048,7 @@ int main (int argc, char **argv) {
                 server_graceful_state(srv);
             }
 
+            if (NULL == srv->conns) rc = 0;
             if (2 == graceful_shutdown) { /* value 2 indicates idle timeout */
                 log_error(srv->errh, __FILE__, __LINE__,
                   "server stopped after idle timeout");
@@ -1985,6 +2066,7 @@ int main (int argc, char **argv) {
         }
 
         /* clean-up */
+        chunkqueue_internal_pipes(0);
         remove_pid_file(srv);
         config_log_error_close(srv);
         if (graceful_restart)
